package de.upb.infexba.ae.entity;

import java.util.ArrayList;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.uima.UimaContext;
import org.apache.uima.cas.FSIterator;
import org.apache.uima.jcas.JCas;
import org.apache.uima.jcas.cas.BooleanArray;
import org.apache.uima.jcas.tcas.Annotation;
import org.apache.uima.resource.ResourceInitializationException;

import de.upb.infexba.MonetaryExpression;
import de.upb.infexba.TimeExpression;
import de.upb.infexba.ae.InfexBAAnalysisEngine;
import de.upb.infexba.linguistic.MonetarySubexpression;
import de.upb.infexba.linguistic.RegexMoney;
import de.upb.infexba.linguistic.SubexpressionType;

/**
 * <p>
 * This analysis engine represents a pattern matcher for finding and 
 * automatically annotating monetary expressions in a German text.
 * <p>
 * The pattern matcher uses the regular expressions given in 
 * {@link RegexMoney} to find all monetary subexpressions.
 * <p>
 * Based on these subexpressions, the complete monetary expressions are
 * determined by finding sequences of subexpressions that fulfill certain
 * criteria. Currently, the following rules are used:
 * <ul>
 *   <li><b>IF</b> there are more than two characters between <code>subexpressions(cur)</code>
 *       and <code>subexpressions(cur-1)</code>
 *       <b>THEN</b> <code>subexpressions(cur)</code> is the start of the next monetary 
 *       expression.</li>
 *   <li>...</li>
 * </ul>
 * 
 * @see {@link MonetarySubexpression} 
 * @author hwachsmuth
 *
 */
public class MoneyExpMatcher extends InfexBAAnalysisEngine {
	
	/**
	 * The Regex pattern for monetary expressions
	 */
	private Pattern pattern;
	
	@Override
	public void initialize(UimaContext aContext) throws ResourceInitializationException {
		super.initialize(aContext);
		try {
			// Compile pattern
			pattern = Pattern.compile(RegexMoney.MONEY_SUBEXPRESSION);
		}
		catch (Exception e){
			e.printStackTrace();
		}
	}	
	
	@Override
	protected void processDocument(JCas jcas) {
		String text = jcas.getDocumentText();		
		// 1st find subexpressions, 2nd find complete expressions
		List<MonetarySubexpression> subexpressions = findSubexpressions(text, jcas);
		findCompleteExpressions(jcas, subexpressions);
	}
	
	/**
	 * Computes and returns all monetary subexpressions found in the <code>text</code>
	 * matched against the given <code>regex</code>.
	 * 
	 * @param text The text to be analysed
	 * @param regex The regular expression to be matched against the text
	 * @param jcas The JCas object
	 * @return The monetary subexpressions found in the text
	 */
	private List<MonetarySubexpression> findSubexpressions(String text, JCas jcas){
		String textToMatch = text;
		Matcher matcher = pattern.matcher(textToMatch);

		//  For every match add the corresponding subexpression to the list
		// together with its type and its according indices.
		List<MonetarySubexpression> subexpressions = new ArrayList<MonetarySubexpression>();
		int index = 0;
		while (matcher.find(index)){			
			// Check if time information intersects the found string
			TimeExpression timeExp = null;
			Annotation tmp = new Annotation(jcas, matcher.start()+1, matcher.end()-1);
			FSIterator<Annotation> timeIter = jcas.getAnnotationIndex(TimeExpression.type).iterator();	
			while(timeIter.hasNext()){
				 TimeExpression next = (TimeExpression) timeIter.next();
				 if ((next.getEnd() >= tmp.getBegin() && next.getEnd() < tmp.getEnd())
						 |(next.getBegin() >= tmp.getBegin() && next.getBegin() < tmp.getEnd())){
					 if (next.getBegin() == 0 || jcas.getDocumentText().charAt(next.getBegin()-1) != '(')
						 timeExp = next;
				 }
			}
			if (timeExp != null) index = timeExp.getEnd();
			
			// No intersection with time information
			else{			
				// Notice that first character isn't taken since it's only used in the regex
				// for findinding the boundary
				String subexpression = matcher.group().substring(1);
				SubexpressionType type = findSubexpressionType(subexpression);
				subexpressions.add(new MonetarySubexpression(
						matcher.start()+1, matcher.end()-1, subexpression, type));
				index = matcher.end()-1;
			}
		}
		return subexpressions;
	}
	
	/**
	 * Determines the type of the given subexpression as defined in the
	 * enumeration {@link SubexpressionType}.
	 * @param subexpression The subexpression to be analysed
	 * @return The type of the subexpression
	 */
	private SubexpressionType findSubexpressionType(String subexpression){
		for (SubexpressionType type : SubexpressionType.values()) {
			int l = type.getPrep().length();
			// Avoid Ouf-of-bounds-exception
			if (subexpression.length() > l
					// Compare leading prep
					&& subexpression.substring(0, l).equals(type.getPrep())
					&& Character.isWhitespace(subexpression.charAt(l))){
				return type;
			}
		}
		// Subexpression without leading prep
		return SubexpressionType.NON;
	}
	
	/**
	 * Computes all the monetary expressions to be annotated in the document text
	 * given in <code>jcas</code> and creates the corresponding annotations. A
	 * monetary expression is a sequence of monetary subexpressions with some
	 * restrictions. These are checked in <code>isEndOfMonetaryExpression</code>. 
	 * 
	 * @param jcas The JCas object
	 * @param subexpressions All the monetary subexpressions in the document text
	 */
	private void findCompleteExpressions(JCas jcas, 
			List<MonetarySubexpression> subexpressions){
		int start = 0;
		// Check for all subexpressions if they belong to a bigger expression
		while (start < subexpressions.size()){
			int cur = start+1;
			// Find end of complete expression
			while (cur < subexpressions.size()){
				if (isStartOfNewMonetaryExpression(subexpressions, jcas.getDocumentText(), start, cur))
					break;
				cur++;
			}		
			// Monetary expression spans from start to cur-1 
			int begin = subexpressions.get(start).getBegin();
			int end = subexpressions.get(cur-1).getEnd();
			MonetaryExpression monetaryExpression = new MonetaryExpression(jcas, begin, end);
			monetaryExpression.setTypes(this.getSubexpressionTypes(jcas, subexpressions, begin, end));
			monetaryExpression.addToIndexes();
			// Next expressions starts at cur
			start=cur;
		}
	}
	
	/**
	 * Returns all the money subexpression types that appear in the span defined by
	 * begin and end and sets the corresponding values in the returned array to true.
	 * 
	 * @param jcas The JCas object
	 * @param subexpressions The subexpressions
	 * @param begin The begin index of the span
	 * @param end The end index of the span
	 * @return The array with an indicator for every type
	 */
	private BooleanArray getSubexpressionTypes(JCas jcas, List<MonetarySubexpression> subexpressions,
			int begin, int end){
		BooleanArray types = new BooleanArray(jcas, SubexpressionType.values().length);
		for (int i = 0; i < types.size(); i++)
			types.set(i, false);
		// Set those types to true which appear in span
		for (MonetarySubexpression subexpression : subexpressions) {
			if (subexpression.getBegin() >= begin && subexpression.getEnd() <= end){
				SubexpressionType type = subexpression.getType();
				// There is a fixed order: alphabetically
				if (type == SubexpressionType.AFTER) types.set(0, true);
				if (type == SubexpressionType.AT) types.set(1, true);
				if (type == SubexpressionType.BETWEEN) types.set(2, true);
				if (type == SubexpressionType.BY) types.set(3, true);
				if (type == SubexpressionType.NON) types.set(4, true);
				if (type == SubexpressionType.OF_OR_FROM) types.set(5, true);
				if (type == SubexpressionType.TO) types.set(6, true);
				if (type == SubexpressionType.WITH) types.set(7, true);
			}
		}
		return types;
	}
	
	/**
	 * Determine whether the monetary subexpression given at index <code>cur</code> in
	 * the list <code>subexpressions</code> is the start subexpression of the next monetary
	 * expression. Currently, the following rules are used:
	 * <ul>
	 *   <li><b>IF</b> there are more than two characters between <code>subexpressions(cur)</code>
	 *       and <code>subexpressions(cur-1)</code>
	 *       <b>THEN</b> <code>subexpressions(cur)</code> is the start of the next monetary 
	 *       expression.</li>
	 *   <li>...</li>
	 * </ul>
	 * 
	 * @param subexpressions The subexpressions to be analyzed
	 * @param start The start list index of the current monetary expression
	 * @param cur The list index of the subexpression to be checked
	 * @return <code>true</code> iff. the checked subexpression is the start subexpression
	 * of a new monetary expression.
	 */
	private boolean isStartOfNewMonetaryExpression(List<MonetarySubexpression> subexpressions,
			String text, int start, int cur){
		// Seperated by comma
		int i = subexpressions.get(cur-1).getEnd();
		if (text.charAt(i) == ','){
			i++;
			while (Character.isWhitespace(text.charAt(i)))
				i++;
			if (i == subexpressions.get(cur).getBegin()) return false;
		}
		// A maximum of two characters may be between two connected subexpressions
		if (subexpressions.get(cur).getBegin()-subexpressions.get(cur-1).getEnd() > 1)
			return true;
		return false;
	}
	
	
	
	
	
	
	
	
	/**
	 * Computes and returns all the money subexpression types that appear in the span defined by
	 * begin and end and sets the corresponding values in the returned array to true.
	 * 
	 * @param jcas The JCas object
	 * @param subexpressions The subexpressions
	 * @param begin The begin index of the span
	 * @param end The end index of the span
	 * @return The array with an indicator for every type
	 */
	public static BooleanArray computeSubexpressionTypes(JCas jcas, int begin, int end){
		// Compute types
		List<MonetarySubexpression> subexpressions = 
			computeSubexpressions(jcas.getDocumentText());//.substring(begin, end));
		BooleanArray types = new BooleanArray(jcas, SubexpressionType.values().length);
		for (int i = 0; i < types.size(); i++)
			types.set(i, false);
		// Set those types to true which appear in span
		for (MonetarySubexpression subexpression : subexpressions) {
			if (subexpression.getBegin() >= begin && subexpression.getEnd() <= end){
				SubexpressionType type = subexpression.getType();
				// There is a fixed order: alphabetically
				if (type == SubexpressionType.AFTER) types.set(0, true);
				if (type == SubexpressionType.AT) types.set(1, true);
				if (type == SubexpressionType.BETWEEN) types.set(2, true);
				if (type == SubexpressionType.BY) types.set(3, true);
				if (type == SubexpressionType.NON) types.set(4, true);
				if (type == SubexpressionType.OF_OR_FROM) types.set(5, true);
				if (type == SubexpressionType.TO) types.set(6, true);
				if (type == SubexpressionType.WITH) types.set(7, true);
			}
		}
		return types;
	}
	
	/**
	 * Computes and returns all monetary subexpressions found in the <code>text</code>
	 * matched against the given <code>regex</code>.
	 * 
	 * @param text The text to be analysed
	 * @param regex The regular expression to be matched against the text
	 * @param jcas The JCas object
	 * @return The monetary subexpressions found in the text
	 */
	private static List<MonetarySubexpression> computeSubexpressions(String text){
		Matcher matcher = (Pattern.compile(RegexMoney.MONEY_SUBEXPRESSION)).matcher(text);

		//  For every match add the corresponding subexpression to the list
		// together with its type and its according indices.
		List<MonetarySubexpression> subexpressions = new ArrayList<MonetarySubexpression>();
		int index = 0;
		while (matcher.find(index)){			
			// Notice that first character isn't taken since it's only used in the regex
			// for findinding the boundary
			String subexpression = matcher.group().substring(1);
			SubexpressionType type = computeSubexpressionType(subexpression);
			subexpressions.add(new MonetarySubexpression(
					matcher.start()+1, matcher.end()-1, subexpression, type));
			index = matcher.end()-1;
		}
		return subexpressions;
	}
	
	/**
	 * Determines the type of the given subexpression as defined in the
	 * enumeration {@link SubexpressionType}.
	 * @param subexpression The subexpression to be analysed
	 * @return The type of the subexpression
	 */
	private static SubexpressionType computeSubexpressionType(String subexpression){
		for (SubexpressionType type : SubexpressionType.values()) {
			int l = type.getPrep().length();
			// Avoid Ouf-of-bounds-exception
			if (subexpression.length() > l
					// Compare leading prep
					&& subexpression.substring(0, l).equals(type.getPrep())
					&& Character.isWhitespace(subexpression.charAt(l))){
				return type;
			}
		}
		// Subexpression without leading prep
		return SubexpressionType.NON;
	}
	
}
