package de.upb.infexba.ae.preprocessing;

import is2.data.SentenceData09;
import is2.lemmatizer3.Lemmatizer;
import is2.lemmatizer3.Options;

import java.io.IOException;
import java.util.List;

import org.apache.uima.UimaContext;
import org.apache.uima.cas.FSIterator;
import org.apache.uima.jcas.JCas;
import org.apache.uima.jcas.tcas.Annotation;
import org.apache.uima.resource.ResourceInitializationException;

import de.upb.infexba.Sentence;
import de.upb.infexba.Token;
import de.upb.infexba.ae.InfexBAAnalysisEngine;
import de.upb.infexba.util.AnnoTools;

/**
 * Analysis engine that wraps the lemmatizer included in the 
 * <A HREF="http://code.google.com/p/mate-tools/" target="_blank">mate-tools</A> of 
 * Bernd Bohnet, University of Stuttgart. 
 * Given token and sentence annotations, this analysis engine computes the lemma for 
 * each token.
 * It is language independent that means no lexicon or morphology is used which would
 * provide extra data. Everything is aquired during the training and no manual langauge 
 * dependent adaption is needed.<BR><BR>
 * 
 * The lemmatizer is based on a statistical model. The location of this model has
 * to be set via the parameter <CODE>MateToolsLemmatizerModel</CODE>.
 * 
 * @author hwachsmuth, 10/15/2010
 *
 */
public class MateToolsLemmatizerWrapper extends InfexBAAnalysisEngine{
	
	// -----------------------------------------------------------------------------------
	// INITIALIZATION
	// -----------------------------------------------------------------------------------
	
	/**
	 * Config parameter for the model used by the lemmatizer 
	 */
	public static final String PARAM_MODEL = "MateToolsLemmatizerModel";
	
	/**
	 * The mate-tools lemmatizer
	 */
	private Lemmatizer lemmatizer;
	
	/**
	 * Options for the lemmatizer
	 */
	private Options optsLemmatizer;
	
	@Override
	public void initialize(UimaContext aContext) throws ResourceInitializationException {
		super.initialize(aContext);
		// Create the lemmatizer with a model set via parameter
		String model = (String) aContext.getConfigParameterValue(PARAM_MODEL);
		try{ 
			if (super.isActivated()){
				optsLemmatizer = new Options(new String[] {"-model", model});
				lemmatizer = new Lemmatizer(optsLemmatizer);
			}
		} catch(IOException ioe){
			ioe.printStackTrace();
		}
	}

	
	// -----------------------------------------------------------------------------------
	// PROCESSING
	// -----------------------------------------------------------------------------------
	
	@Override
	public void processDocument(JCas jcas){
		this.lemmatizeDocument(jcas);
	}
	
	/**
	 * Lemmatizes all tokens of the document represented in the JCas object.
	 * 
	 * @param jcas The JCas object
	 */
	private void lemmatizeDocument(JCas jcas){
		FSIterator<Annotation> sentenceIter = jcas.getAnnotationIndex(Sentence.type).iterator();
		while (sentenceIter.hasNext())
			this.lemmatizeSentence(jcas, (Sentence) sentenceIter.next());
	}

	/**
	 * Lemmatizes all tokens of the given sentence.
	 * 
	 * @param jcas The JCas object
	 * @param sentence The sentence
	 */
	private void lemmatizeSentence(JCas jcas, Sentence sentence){
		// Get tokens and convert to mate-tools structure
		List<Token> tokens = AnnoTools.sentenceToTokens(jcas, sentence);
		List<String> tokenStrings = AnnoTools.tokensToStrings(tokens);
		tokenStrings.add(0, "<root>");
		SentenceData09 sentenceData = new SentenceData09();
		sentenceData.init(tokenStrings.toArray(new String[0]));
		// Lemmatize and relate lemmata to tokens
		lemmatizer.apply(sentenceData);
		for(int i=1; i<sentenceData.lemmas.length; i++){
			tokens.get(i-1).setLemma(sentenceData.lemmas[i]);
//			System.out.println(sentenceData.lemmas[i]);
		}	
//		System.out.println();
	}
}
