package de.upb.infexba.ae.preprocessing;

import java.util.Arrays;
import java.util.LinkedList;
import java.util.List;

import org.apache.commons.pool.BasePoolableObjectFactory;
import org.apache.commons.pool.ObjectPool;
import org.apache.commons.pool.PoolUtils;
import org.apache.commons.pool.impl.GenericObjectPool;
import org.apache.log4j.Logger;
import org.apache.uima.UimaContext;
import org.apache.uima.cas.FSIterator;
import org.apache.uima.cas.text.AnnotationIndex;
import org.apache.uima.jcas.JCas;
import org.apache.uima.jcas.tcas.Annotation;
import org.apache.uima.resource.ResourceInitializationException;

import de.upb.infexba.Sentence;
import de.upb.infexba.Token;
import de.upb.infexba.ae.InfexBAAnalysisEngine;
import de.upb.infexba.ae.preprocessing.TreeTaggerWrapper.TaggedToken;

/**
 * Annotator that uses the external 
 * <a href="http://www.ims.uni-stuttgart.de/projekte/corplex/TreeTagger/" target="_blank">TreeTagger</a> 
 * from the <i>University of Stuttgart</i> to find three tags for each token:
 * <ul>
 *   <li>The part-of-speech tag</li>
 *   <li>The chunk tag</li>
 *   <li>The according lemma</li>
 * </ul>
 * 
 * The annotator needs as input the document text together with token annotations.
 * It writes the tokens to a file (one token per line) and triggers the TreeTagger.
 * The returned tags are then converted into UIMA annotations.
 * 
 * 
 * 
 * @author hwachsmuth
 * @author pprett
 */
public class PartOfSpeechAndChunkTagger extends InfexBAAnalysisEngine{
    
    /**
     * 
     * Factory for TreeTaggerWrappers. 
     *
     * This class populates the taggerPool.
     *
     *
     * @author peter.prettenhofer@uni-weimar.de
     * @version aitools 2.0
     *
     * Created on Oct 27, 2010
     *
     * $Id$
     */
    public class TaggerFactory extends BasePoolableObjectFactory { 
        // for makeObject we'll simply return a new buffer 
        @Override
        public Object makeObject() { 
            return new TreeTaggerWrapper(treeTaggerDirectory);
        } 
        
        /**
         * Free resources properly. 
         */
        @Override
        public void destroyObject(Object object) { 
            if (object instanceof TreeTaggerWrapper) {
                TreeTaggerWrapper tagger = (TreeTaggerWrapper) object;
                tagger.close();
            } else {
                throw new IllegalArgumentException("Pool must only contain TreeTaggerWrapper objects.");
            }
        }
    }

	
	public static final String PARAM_TREETAGGER_DIR = "TreeTaggerDirectory";
	public static final String PARAM_INITIAL_CAPACITY= "InitialCapacity";
	public static final String PARAM_MAX_ACTIVE = "MaxActive";
	
	private static final Logger LOGGER = Logger.getLogger(PartOfSpeechAndChunkTagger.class);

    private int initialCapacity = 1;
    private int maxActive = 4;
	ObjectPool taggerPool;
	//Timer timer;
	
	private String treeTaggerDirectory;
	
	
	// -------------------------------------------------------------------
	// UIMA INTERFACE METHODS
	// -------------------------------------------------------------------
	
	@Override
	public void initialize(UimaContext aContext) throws ResourceInitializationException {
		super.initialize(aContext);
		if (super.isActivated()){
		    treeTaggerDirectory = (String) aContext.getConfigParameterValue(PARAM_TREETAGGER_DIR);
		    initialCapacity = (Integer) aContext.getConfigParameterValue(PARAM_INITIAL_CAPACITY);
		    maxActive = (Integer) aContext.getConfigParameterValue(PARAM_MAX_ACTIVE);
		    taggerPool = PoolUtils.synchronizedPool(new GenericObjectPool(new TaggerFactory(), maxActive,  GenericObjectPool.WHEN_EXHAUSTED_BLOCK, -1));
		    try {
		    for (int i = 0; i < initialCapacity; i++) {
		        taggerPool.addObject();
		    }
		    } catch(Exception e) {
		        LOGGER.error(e);
		        throw new RuntimeException("cannot populate tagger pool.", e);
		    }
		    
		}
//// Debug for object pool status.
//		timer = new Timer();
//        timer.schedule(new TimerTask()
//        {
//            @Override
//            public void run()
//            {
//                int nActive = taggerPool.getNumActive();
//                int nIdle = taggerPool.getNumIdle();
//                System.err.printf("pool: %d active, %d idle\n", nActive, nIdle);
//            }
//
//        }, 100, 100);
	}

	@Override
	public void processDocument(JCas jcas){
		try {
	    	String[] tokens = casToTokenSequences(jcas);
	    	if(tokens.length == 0)
	    		return;
	    	TreeTaggerWrapper tagger = (TreeTaggerWrapper) taggerPool.borrowObject();
	    	TaggedToken[] taggedSeq = tagger.chunk(tokens);
	    	taggerPool.returnObject(tagger);
	        createLemmataPOSAndChunkTags(jcas, taggedSeq);
	    } catch (Exception e) {
	        LOGGER.error("Cannot process JCas: " + e.getMessage(), e);
	    }
	}
	
	/**
	 * Iterates over the tokens of the {@link JCas} and accumulates them into an array. 
	 * @param jcas The {@link JCas} to analyze. 
	 * @return An array of String objects. 
	 */
	private String[] casToTokenSequences(JCas jcas) {
	    AnnotationIndex<Annotation> sentenceIndex = jcas.getAnnotationIndex(Sentence.type);
        AnnotationIndex<Annotation> tokenIndex = jcas.getAnnotationIndex(Token.type);

        // iterate over Sentences
        FSIterator<Annotation> sentenceIterator = sentenceIndex.iterator();
        List<String> tokenSeq = new LinkedList<String>();
        while (sentenceIterator.hasNext()) {
          Sentence sentence = (Sentence) sentenceIterator.next();
          // iterate over tokens
          FSIterator<Annotation> tokenIterator = tokenIndex.subiterator(sentence);
          while (tokenIterator.hasNext()) {
            Token token = (Token) tokenIterator.next();
            String word = token.getCoveredText();
            if (word.equals("")) {
                LOGGER.debug("Empty token encountered.");
            } else {
            	if(word.contains("\n")) {
            	    LOGGER.debug(String.format("error: Token contains newline (%s)\n", word));
            		word = word.replace("\n", " ");
            	}
            	tokenSeq.add(word);
            }
          }
          //tokenSeq.add("."); // mark sentence boundaries by blanks.
        }
        return tokenSeq.toArray(new String[0]);
	}
	
	/**
	 * Writes the {@link TaggedToken} sequence to the {@link JCas}. 
	 */
	private void createLemmataPOSAndChunkTags(JCas jcas, TaggedToken[] tags){
		try{
		    LinkedList<TaggedToken> tagstack = new LinkedList<TaggedToken>(Arrays.asList(tags));
		    AnnotationIndex<Annotation> sentenceIndex = jcas.getAnnotationIndex(Sentence.type);
	        AnnotationIndex<Annotation> tokenIndex = jcas.getAnnotationIndex(Token.type);
	        
	        // iterate over Sentences
	        FSIterator<Annotation> sentenceIterator = sentenceIndex.iterator();
	        while (sentenceIterator.hasNext()) {
	          Sentence sentence = (Sentence) sentenceIterator.next();
	          // iterate over tokens
	          FSIterator<Annotation> tokenIterator = tokenIndex.subiterator(sentence);
	          while (tokenIterator.hasNext()) {
	            Token token = (Token) tokenIterator.next();
	            String term = token.getCoveredText();
	            assert term.equals(tagstack.peekFirst().term);
	            if(!term.equals(tagstack.peekFirst().term)) {
	                LOGGER.warn(String.format("in:%s\tout:%s\n", term, tagstack.peekFirst().term));
	            }
	            createAnnotation(jcas, token, tagstack.pop());         
	          }
	        }
		}
		catch (Exception e){
		    LOGGER.error("Cannot write tags to JCas. ", e);
			throw new RuntimeException("Cannot write tags to JCas. ", e);
		}
	}

	/**
	 * Creates the annotations for the {@link Token} corresponding to the information in {@link TaggedToken}.
	 * @param jcas The {@link JCas} which accommodates the {@link Token}.
	 * @param token The {@link Token} to be annotated. 
	 * @param tag The {@link TaggedToken} which carries the annotations. 
	 */
    private void createAnnotation(JCas jcas, Token token, TaggedToken tag)
    {
        if(tag.lemma!=null && !tag.lemma.equals("")) {
            token.setLemma(tag.lemma);
        }
        token.setPos(tag.pos);
        token.setChunk(tag.chunk);
    }	
    
    @Override
    public void destroy() {
        super.destroy();
//// Debug for object pool status.
// timer.cancel();
        try {
        	if (super.isActivated()) {
        	    LOGGER.info("Closing Pool - destroying all TreeTaggerWrapper objects.");
        	    taggerPool.close();
        	}
        }
        catch (Exception e) {
            LOGGER.error("Error while closing tagger pool.", e);
            throw new RuntimeException("Error while closing tagger pool.", e);
        }
    }
}
