package de.upb.infexba.application;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.List;
import java.util.TimeZone;

import org.apache.uima.UIMAFramework;
import org.apache.uima.analysis_engine.AnalysisEngine;
import org.apache.uima.analysis_engine.AnalysisEngineProcessException;
import org.apache.uima.cas.CAS;
import org.apache.uima.cas.CASException;
import org.apache.uima.cas.FSIterator;
import org.apache.uima.collection.CollectionException;
import org.apache.uima.collection.CollectionReader;
import org.apache.uima.jcas.JCas;
import org.apache.uima.jcas.tcas.Annotation;
import org.apache.uima.resource.ResourceConfigurationException;
import org.apache.uima.resource.ResourceInitializationException;
import org.apache.uima.resource.ResourceSpecifier;
import org.apache.uima.util.InvalidXMLException;
import org.apache.uima.util.XMLInputSource;

import de.upb.infexba.Genre;
import de.upb.infexba.cr.XmiCollectionReader;
import de.upb.infexba.ml.GeneralFeatureARFFReader;
import de.upb.infexba.ml.GeneralFeatureARFFWriter;
import de.upb.infexba.ml.IntentionFeatureGenerator;
import de.upb.infexba.ml.feature.FeatureHandler;

/**
 * Class for the generation of feature sets and according feature values for the statement type classification problem, i.e. is a given statement a forecast or
 * a declaration.
 * <p>
 * If the parameter <CODE>split</CODE> is set to <CODE>"training"</CODE> then, first, a feature set is created based on the given training input directory,
 * which is then balanced if specified.<BR>
 * The feature generation and computation is done in the class {@link de.upb.infexba.ml.TypeFeatureGenerator}. A WEKA feature ARFF file is then generated at the
 * specified location in the way specified in {@link de.upb.infexba.ml.GeneralFeatureARFFWriter}.
 * <p>
 * If the parameter is set to <CODE>"validation"</CODE> or <CODE>"test"</CODE> only the generation and computation of feature is done for the corresponding
 * input directories. Therefore, the feature set from the training file is loaded beforehand. Notice that this file needs to be in the home directory of the
 * project.
 * <p>
 * The input directory may be set via the first argument of the main method.
 * 
 * @author hwachsmuth, 2010/10/25
 * 
 */
public class GenerationOfIntentionFeatures {

	// -------------------------------------------------------------------
	// FIXED FILE PATHS AND NAMES
	// -------------------------------------------------------------------

	/**
	 * Location of collection reader
	 */
	private static String collectionReaderDesc = "descriptors/collection/RevenueCorpusReader.xml";

	/**
	 * Location of aggregate revenue statement extractor
	 */
	private static String analysisEngineDesc_fast = "descriptors/aggregate/AggregatePreprocessing.xml";

	private static String analysisEngineDesc_mate = "descriptors/aggregate/AggregateMatePreprocessing.xml";

	/**
	 * Path of input directory (split will be added in main)
	 */
	private static String inputDir = "corpora/lfa-corpus/";

	private static String musicFolder = "music_split/";

	private static String smartphoneFolder = "smartphone_split/";

	/**
	 * Path of the directory where the .arff files are stored
	 */
	private static String arffDir = "arff/";

	/**
	 * The prefix of all type feature files
	 */
	private static final String FILE_PREFIX = "dataset_intention_";

	/**
	 * The suffix of all type feature files
	 */
	private static final String FILE_SUFFIX = ".arff";

	// -------------------------------------------------------------------
	// OPTIONS
	// -------------------------------------------------------------------

	/**
	 * Possible corpora.
	 */
	public static final boolean MUSIC = true, SMARTPHONE = false;

	/**
	 * determines whether music or smartphone corpus shall be used
	 */
	public static boolean corpus = MUSIC;

	/**
	 * Possible modes.
	 */
	public static final byte MODE_SAMPLE = -1, MODE_TRAIN = 0, MODE_VALIDATE = 1, MODE_TEST = 2, MODE_TRAIN_AND_VALIDATE = 3,
			MODE_TRAIN_AND_VALIDATE_AND_TEST = 4;

	/**
	 * Specifies which mode shall be used.
	 */
	public static byte mode = GenerationOfIntentionFeatures.MODE_TEST;

	/**
	 * If you want to create a validation/test set (i.e. mode is set to MODE_VALIDATE or MODE_TEST) w.r.t. a certain feature-set given in an .ARFF-file, specify it here.
	 */
	public static String arffFile = "dataset_intention_05-25 11 49 00_smartphone_split-training.arff";

	/**
	 * Possible normalization modes.
	 */
	static final short NORM_NONE = 0, NORM_MY = 1, NORM_CALCULATED_MAXIMA = 2, NORM_CALCULATED_AVERAGE = 3;

	/**
	 * determines which kind of normalization shall be used. Note: atm only used for POS-tags!
	 */
	public static short normalizationMode = GenerationOfIntentionFeatures.NORM_CALCULATED_AVERAGE;

	/**
	 * Seed used for over-/under-sampling.
	 */
	static long seed = 321353169; // standard: 321353169

	/**
	 * <CODE>true</CODE> if and only if the set of the features to be created shall be balanced via under-sampling. Else over-sampling is used. Balancing only
	 * works for training sets, in general.
	 */
	private static boolean undersampling = false;

	/**
	 * <CODE>true</CODE> if and only if you want to use the MATE-Lemmatizer/POS-Tagger instead.
	 */
	private static boolean use_mate = false;

	// -------------------------------------------------------------------
	// MAIN METHOD – GENERATION METHOD
	// -------------------------------------------------------------------

	/**
	 * The main method executes the evaluation process. That is, it computes recall, precision and value that refer to these for the statements found using the
	 * extraction algorithm.
	 * <p>
	 * Additionally, the recall of the found sentence, time and money expressions is computed.
	 * <p>
	 * Both are printed to screen.
	 * 
	 * @param args Not used
	 */
	public static void main(String[] args) {

		// data-sets?

		String SAMPLE = "music_split/sample";
		String TRAINING = "";
		String VALIDATION = "";
		String TEST = "";
		boolean get_first_annotation = false;
		String folder = "";
		if (corpus) {
			// music -> always use first annotation
			get_first_annotation = true;
			folder = musicFolder;
		} else {
			// music -> always use second annotation
			get_first_annotation = false;
			folder = smartphoneFolder;
		}
		TRAINING = folder + "training";
		VALIDATION = folder + "validation";
		TEST = folder + "test";

		// normalization mode?

		if (normalizationMode == NORM_CALCULATED_MAXIMA || normalizationMode == NORM_CALCULATED_AVERAGE)
			if (mode == MODE_VALIDATE || mode == MODE_TEST) {
				if (arffFile.contains("music")) {
					if (normalizationMode == NORM_CALCULATED_MAXIMA)
						normalizationMode = FeatureHandler.CALCULATED_MAXIMA_MUSIC;
					else
						normalizationMode = FeatureHandler.CALCULATED_AVERAGE_MUSIC;
				} else {
					if (normalizationMode == NORM_CALCULATED_MAXIMA)
						normalizationMode = FeatureHandler.CALCULATED_MAXIMA_SMARTPHONE;
					else
						normalizationMode = FeatureHandler.CALCULATED_AVERAGE_SMARTPHONE;
				}
			} else {
				if (corpus) {
					if (normalizationMode == NORM_CALCULATED_MAXIMA)
						normalizationMode = FeatureHandler.CALCULATED_MAXIMA_MUSIC;
					else
						normalizationMode = FeatureHandler.CALCULATED_AVERAGE_MUSIC;
				} else {
					if (normalizationMode == NORM_CALCULATED_MAXIMA)
						normalizationMode = FeatureHandler.CALCULATED_MAXIMA_SMARTPHONE;
					else
						normalizationMode = FeatureHandler.CALCULATED_AVERAGE_SMARTPHONE;
				}
			}
		else if (normalizationMode == NORM_MY)
			normalizationMode = FeatureHandler.MY_NORMALIZATION;
		else if (normalizationMode == NORM_NONE)
			normalizationMode = FeatureHandler.NO_FACTORS;

		System.out.println("NormalizationMode set to " + normalizationMode);

		// mode -> data-set

		String dataset = TRAINING;
		int rounds = 1;
		switch (mode) {
		case MODE_SAMPLE:
			dataset = SAMPLE;
			break;
		case MODE_TRAIN:
			dataset = TRAINING;
			break;
		case MODE_VALIDATE:
			dataset = VALIDATION;
			break;
		case MODE_TEST:
			dataset = TEST;
			break;
		case MODE_TRAIN_AND_VALIDATE:
			dataset = TRAINING;
			rounds = 2;
			break;
		case MODE_TRAIN_AND_VALIDATE_AND_TEST:
			dataset = TRAINING;
			rounds = 3;
			break;
		}

		if ((mode == MODE_VALIDATE || mode == MODE_TEST) && arffFile.isEmpty()) {
			System.err.println("Please specify an arff-file.");
			return;
		}

		XMLInputSource xmlInputSource;
		ResourceSpecifier specifier;
		AnalysisEngine analysisEngine;
		try {

			// ~~~~~~~~ Create analysis engine ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

			String analysisEngineDesc = analysisEngineDesc_fast;
			if (use_mate)
				analysisEngineDesc = analysisEngineDesc_mate;

			xmlInputSource = new XMLInputSource(analysisEngineDesc);
			specifier = UIMAFramework.getXMLParser().parseResourceSpecifier(xmlInputSource);
			analysisEngine = UIMAFramework.produceAnalysisEngine(specifier);

			if (use_mate) {
				analysisEngine.setConfigParameterValue("MateToolsLemmatizerWrapper/Activated", "true");
				analysisEngine.setConfigParameterValue("MateToolsPOSTaggerWrapper/Activated", "true");
			}

			analysisEngine.setConfigParameterValue("RBTokenizer/Activated", "true");
			analysisEngine.setConfigParameterValue("RBSentenceSplitter/Activated", "true");
			analysisEngine.setConfigParameterValue("PartOfSpeechAndChunkTagger/Activated", "true");

			analysisEngine.setConfigParameterValue("TimeExpMatcher/Activated", "true");
			analysisEngine.setConfigParameterValue("MoneyExpMatcher/Activated", "true");

			analysisEngine.reconfigure();

			// ~~~~~~~~ Create feature generator ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

			IntentionFeatureGenerator generator = new IntentionFeatureGenerator();
			generator.setGetFirstAnnotation(get_first_annotation);
			generator.setSeed(seed);
			generator.setNormalizationMode(normalizationMode);

			// ~~~~~~~~ Create sets for training, validation,.. in multiple rounds ~~~~~~~~~~~~~

			for (int r = 0; r < rounds; r++) {

				String curInputDir = inputDir + dataset + "/";

				// ~~~~~~~~ Collection-Reader ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

				// Get the collection to be iterated over
				xmlInputSource = new XMLInputSource(collectionReaderDesc);
				specifier = UIMAFramework.getXMLParser().parseResourceSpecifier(xmlInputSource);
				CollectionReader collectionReader = UIMAFramework.produceCollectionReader(specifier);
				collectionReader.setConfigParameterValue(XmiCollectionReader.PARAM_INPUTDIR, curInputDir);
				collectionReader.reconfigure();

				// ~~~~~~~~ Initialize the feature set... ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
				int number_of_trainingfiles = new File(curInputDir).list().length;
				String info = "";
				if (dataset.equals(TRAINING) || dataset.equals(SAMPLE)) {
					System.out.print("Initialize feature set...");
					int doc = 1;
					CAS aCAS = analysisEngine.newCAS();

					int no_classification = 0;
					while (collectionReader.hasNext()) {
						doc++;
						if (doc % 10 == 0)
							System.out.print(".");
						collectionReader.getNext(aCAS);
						JCas jcas = aCAS.getJCas();

						if (has_classification(jcas)) {
							// Preprocess documents
							analysisEngine.process(jcas);

							// Look for new features
							generator.initializeFeatureSet(jcas, number_of_trainingfiles);

						} else {
							no_classification++;
						}
					}
					System.out.print("skipped " + no_classification + " file(s) without classification ...");
					System.out.println("finished");

					// Reduce feature set
					info = generator.reduceFeatureSet();
				}

				// ~~~~~~~~ ... or load it from training file ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

				else {
					System.out.print("\nLoad feature set from file " + arffFile + "...");

					GeneralFeatureARFFReader reader = new GeneralFeatureARFFReader();
					List<String> featureNames = reader.readFeatureSetFromARFF(arffDir + arffFile);
					generator.setFeatureSet(featureNames);

					System.out.println("finished");
				}

				// ~~~~~~~~ Compute the features ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

				System.out.print("\nCompute features...");
				collectionReader.reconfigure();
				int doc = 1;
				CAS aCAS = analysisEngine.newCAS();
				while (collectionReader.hasNext()) {
					doc++;
					if (doc % 10 == 0)
						System.out.print(".");
					collectionReader.getNext(aCAS);
					JCas jcas = aCAS.getJCas();

					if (has_classification(jcas)) {
						// preprocess documents
						analysisEngine.process(jcas);

						// Generate feature values for the given JCas
						generator.computeFeatures(jcas);
					}
				}
				System.out.println("finished");

				// ~~~~~~~~ Balance the training-set ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

				// only balance the training-set

				if (dataset.equals(TRAINING))
					generator.balance(undersampling);

				// ~~~~~~~~ Output ARFF-file and protocol ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

				SimpleDateFormat df = new SimpleDateFormat("MM-dd HH mm ss");
				df.setTimeZone(TimeZone.getDefault());
				String name = dataset.replace('/', '-');

				// Create an protocol file
				if (r == 0) {
					File protocol = new File(arffDir + FILE_PREFIX + df.format(new Date()) + "_protocol.txt");
					Writer out = new OutputStreamWriter(new FileOutputStream(protocol));
					try {
						out.write("Result (cross-valid/valid/smartphone): \n\n" + "Notiz: \n\n" + "used MATE: " + use_mate + "\n\nundersampling: "
								+ undersampling + "\n\n" + info);
					} finally {
						out.close();
					}
				}

				// Write to file
				GeneralFeatureARFFWriter writer = new GeneralFeatureARFFWriter("Intentions of documents", "intention", "{"
						+ IntentionFeatureGenerator.COMMERCIAL + "," + IntentionFeatureGenerator.INFORMATIONAL + "," + IntentionFeatureGenerator.PERSONAL + "}");
				String tmpOutputFile = FILE_PREFIX + df.format(new Date()) + "_" + name + FILE_SUFFIX;
				writer.writeFeaturesToARFF(generator.getFeatureNames(), generator.getFeatures(), arffDir + tmpOutputFile);

				// Set outputFile for next validation/test-round
				if (r == 0)
					arffFile = tmpOutputFile;

				// "Restart" the generator (i.e. old computed features are deleted)
				generator.clear();

				// set split for the next round: round 0 -> TRAINING, round 1 -> VALIDATION, round 2 -> 2nd VALIDATION
				// or TEST
				if (r == 0)
					dataset = VALIDATION;
				else if (r == 1)
					dataset = TEST;

			} // end for

			analysisEngine.destroy();

		} catch (CASException e) {
			e.printStackTrace();
		} catch (CollectionException e) {
			e.printStackTrace();
		} catch (AnalysisEngineProcessException e) {
			e.printStackTrace();
		} catch (InvalidXMLException e1) {
			e1.printStackTrace();
		} catch (ResourceInitializationException e1) {
			e1.printStackTrace();
		} catch (ResourceConfigurationException e1) {
			e1.printStackTrace();
		} catch (IOException e1) {
			e1.printStackTrace();
		}

	}

	/**
	 * Returns the first Classification of the JCas Object saved in its annotation.
	 * 
	 * @param jcas
	 * @return
	 */
	private static boolean has_classification(JCas jcas) {
		// Get the classification from the jCas object.
		boolean has_classification = false;
		String classification = "NONE";
		FSIterator<Annotation> genreIter = jcas.getAnnotationIndex(Genre.type).iterator();
		if (genreIter.hasNext()) {
			Annotation a = genreIter.next();
			if (a.getTypeIndexID() == Genre.type) {
				classification = ((Genre) a).getGenreClass();
				has_classification = classification.equals(IntentionFeatureGenerator.PERSONAL) || classification.equals(IntentionFeatureGenerator.COMMERCIAL)
						|| classification.equals(IntentionFeatureGenerator.INFORMATIONAL);
			}
		}
		return has_classification;
	}
}
