/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package de.upb.infexba.application;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.uima.UIMAFramework;
import org.apache.uima.analysis_engine.AnalysisEngine;
import org.apache.uima.cas.CAS;
import org.apache.uima.cas.FSIterator;
import org.apache.uima.cas.Feature;
import org.apache.uima.cas.FeatureStructure;
import org.apache.uima.cas.Type;
import org.apache.uima.cas.text.AnnotationFS;
import org.apache.uima.collection.CollectionReader;
import org.apache.uima.resource.ResourceSpecifier;
import org.apache.uima.util.XMLInputSource;

import de.upb.infexba.cr.XmiCollectionReader;

/**
 * Counts various numbers concerning the annotated classes of the input documents, i.e., the relevance, genre and the sentiment of the documents. Prints these
 * numbers to screen.
 * 
 * <p>
 * The application may take two arguments:
 * <ol type="1">
 * <li>The input directory containing the files to be processed. If you don't specify this directory, corpora/sentiment/smartphones is used.</li>
 * <li>The path to an XML descriptor for the type system to be used. If you don't specify this path, descriptors/infexbaTypeSystem.xml is used.</li>
 * </ol>
 * 
 * @author hwachsmuth
 * 
 */
public class KappaStatistic {

	// ==========================================================================
	// PATHS
	// ==========================================================================

	/**
	 * Location of collection reader
	 */
	private static String collectionReaderDesc = "descriptors/collection/SentimentCorpusReader.xml";

	/**
	 * Location of type system
	 */
	private static String typeSystemDesc = "descriptors/infexbaTypeSystem.xml";

	/**
	 * Location of analysis engine
	 */
	private static String analysisEngineDesc = "descriptors/collection/DummyAnalysisEngine.xml";

	/**
	 * Path of input directory
	 */
	private static String inputDir =  "corpora/music_double/"; //"corpora/smartphone_double/"; // 

	// ==========================================================================
	// COUNTERS
	// ==========================================================================

	/**
	 * Map that stores the number of occurrences of each of the four genre classes in the given corpus.
	 */
	private static Map<String, Integer> genreclassMap;

	/**
	 * Map that stores the number of occurrences of documents in the given corpus that contain 0, 1 or 2 metadata tags, respectively.
	 * 
	 */
	private static Map<String, Integer> metadataMap;

	// ==========================================================================
	// TYPE NAMES
	// ==========================================================================

	/**
	 * Name of the type of entities
	 */
	private static final String GENRE = "de.upb.infexba.Genre";

	/**
	 * Name of the type of declarations
	 */
	private static final String METADATA = "de.upb.infexba.Metadata";

	/**
	 * The name of the intention class commercial
	 */
	public static final String COMMERCIAL = "commercial";

	/**
	 * The name of the intention class informational
	 */
	public static final String INFORMATIONAL = "informational";

	/**
	 * The name of the intention class personal
	 */
	public static final String PERSONAL = "personal";

	// ==========================================================================
	// COUNTING
	// ==========================================================================

	/**
	 * Main method iterates over all documents in the collection and sums up the number of included annotations.
	 * 
	 * @param args See class description
	 */
	public static void main(String[] args) {
		// Set arguments if existing
		if (args.length > 0)
			inputDir = args[0];
		if (args.length > 1)
			typeSystemDesc = args[1];

		try {
			// Get type system for initializing the counter map
			XMLInputSource in = new XMLInputSource(typeSystemDesc);

			// Initialize entries of maps for counting
			genreclassMap = new HashMap<String, Integer>();
			metadataMap = new HashMap<String, Integer>();

			// Get the collection to be iterated over
			in = new XMLInputSource(collectionReaderDesc);
			ResourceSpecifier specifier = UIMAFramework.getXMLParser().parseResourceSpecifier(in);
			CollectionReader cr = UIMAFramework.produceCollectionReader(specifier);
			if (!inputDir.equals("")) {
				cr.setConfigParameterValue(XmiCollectionReader.PARAM_INPUTDIR, inputDir);
				cr.reconfigure();
			}

			// Create an analysis engine, so we can create a CAS
			in = new XMLInputSource(analysisEngineDesc);
			specifier = UIMAFramework.getXMLParser().parseResourceSpecifier(in);
			AnalysisEngine ae = UIMAFramework.produceAnalysisEngine(specifier);
			CAS aCAS = ae.newCAS();

			// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
			// Counting begins here...

			// Fleiss-Kappa statistic
			// explanation: see below

			double fleiss_P_1 = 0;

			// Cohens Kappa statistic
			// Assumption: there are exactly 2 annotators.

			// number of times annotator 1 (or 2) chooses c(ommercial), p(ersonal) or i(nformational)
			double cohen_1_c = 0, cohen_1_p = 0, cohen_1_i = 0;
			double cohen_2_c = 0, cohen_2_p = 0, cohen_2_i = 0;
			// number of times the annotator 1 and 2 agree
			double cohen_agree = 0;
			// and number of times they agree w.r.t. only to a single class (i.e. w.r.t the question whether the the document belongs to the given class or not)
			double cohen_i_agree = 0, cohen_c_agree = 0, cohen_p_agree = 0;

			// helper variables
			int numberOfDocuments = 1;
			int numberOfAnnotations = 0;

			// Iterate over the whole collection
			System.out.print("Read documents...");
			while (cr.hasNext()) {

				cr.getNext(aCAS);

				int metadata = 0;

				// count number of categories that were assigned to the _current_ document
				double c = 0., i = 0., p = 0.;

				// Iterate over all annotations of the document
				FSIterator<AnnotationFS> fsIter = aCAS.getAnnotationIndex().iterator();

				boolean first = true;

				while (fsIter.hasNext()) {

					// Get type of annotation
					Type type = fsIter.get().getType();

					if (type.getName().equals(METADATA)) {
						metadata++;
					} else if (type.getName().equals(GENRE)) {
						numberOfAnnotations++;

						FeatureStructure fs = fsIter.get();
						List<Feature> features = fs.getType().getFeatures();
						Feature feature = features.get(3);
						String className = fs.getFeatureValueAsString(feature);
						if (genreclassMap.containsKey(className)) {
							genreclassMap.put(className, genreclassMap.get(className) + 1);
						} else
							genreclassMap.put(className, 1);

						if (className.equals(COMMERCIAL)) {
							c++;
							if (first)
								cohen_1_c++;
							else
								cohen_2_c++;
						} else if (className.equals(INFORMATIONAL)) {
							i++;
							if (first)
								cohen_1_i++;
							else
								cohen_2_i++;
						} else if (className.equals(PERSONAL)) {
							p++;
							if (first)
								cohen_1_p++;
							else
								cohen_2_p++;
						}
					}

					if (first)
						first = false;

					fsIter.next();
				}

				if (metadataMap.containsKey("" + metadata))
					metadataMap.put("" + metadata, metadataMap.get("" + metadata) + 1);
				else
					metadataMap.put("" + metadata, 1);

				// for Fleiss-Kappa statistic
				fleiss_P_1 += ((p * (p - 1.) + c * (c - 1.) + i * (i - 1.)));

				// for Cohens Kappa statistic
				if (c == 2 || c == 0)
					cohen_c_agree++;
				if (p == 2 || p == 0)
					cohen_p_agree++;
				if (i == 2 || i == 0)
					cohen_i_agree++;
				if (c == 2 || i == 2 || p == 2)
					cohen_agree++;

				numberOfDocuments++;
				if (numberOfDocuments % 20 == 0)
					System.out.print(".");
			}

			System.out.println("number of docs = " + (numberOfDocuments - 1));
			System.out.println("annotations = " + numberOfAnnotations);

			// Counting ends here
			// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
			// Statistics

			double commercial = 0;
			double informational = 0;
			double personal = 0;
			if (genreclassMap.containsKey(COMMERCIAL))
				commercial = genreclassMap.get(COMMERCIAL);
			if (genreclassMap.containsKey(INFORMATIONAL))
				informational = genreclassMap.get(INFORMATIONAL);
			if (genreclassMap.containsKey(PERSONAL))
				personal = genreclassMap.get(PERSONAL);

			// Fleiss Kappa

			// P_1 = Mean value (over all subjects) of the extent to which raters agree for the i-th
			// subject
			fleiss_P_1 = fleiss_P_1 / (3.0 * (numberOfDocuments - 1.));
			// P_2 = Mean value (over all categories j) of the proportion of all assignments which were
			// to the j-th category
			double fleiss_P_2 = (Math.sqrt(1. / commercial) + Math.sqrt(1. / informational) + Math.sqrt(1. / personal))
					/ Math.sqrt(commercial + personal + informational);

			double fleiss_kappa = (fleiss_P_1 - fleiss_P_2) / (1. - fleiss_P_2);

			// Cohens Kappa

			double cohen_P_e = (cohen_1_c * cohen_2_c + cohen_1_p * cohen_2_p + cohen_1_i * cohen_2_i) / Math.sqrt(numberOfDocuments);
			double cohen_P_a = cohen_agree / numberOfDocuments;
			double cohen_kappa = (cohen_P_a - cohen_P_e) / (1 - cohen_P_e);

			double cohen_c_P_e = (cohen_1_c * cohen_2_c + (cohen_1_p + cohen_1_i) * (cohen_2_p + cohen_2_i)) / Math.sqrt(numberOfDocuments);
			double cohen_c_P_a = cohen_c_agree / numberOfDocuments;
			double cohen_c_kappa = (cohen_c_P_a - cohen_c_P_e) / (1 - cohen_c_P_e);

			double cohen_p_P_e = (cohen_1_p * cohen_2_p + (cohen_1_c + cohen_1_i) * (cohen_2_c + cohen_2_i)) / Math.sqrt(numberOfDocuments);
			double cohen_p_P_a = cohen_p_agree / numberOfDocuments;
			double cohen_p_kappa = (cohen_p_P_a - cohen_p_P_e) / (1 - cohen_p_P_e);

			double cohen_i_P_e = (cohen_1_i * cohen_2_i + (cohen_1_c + cohen_1_p) * (cohen_2_c + cohen_2_p)) / Math.sqrt(numberOfDocuments);
			double cohen_i_P_a = cohen_i_agree / numberOfDocuments;
			double cohen_i_kappa = (cohen_i_P_a - cohen_i_P_e) / (1 - cohen_i_P_e);

			// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
			// Print computed numbers
			printStatistics(METADATA, metadataMap);
			printStatistics(GENRE, genreclassMap);

			System.out.println("\n\ninputDir: "+inputDir+"\n");
			
			System.out.println("Fleiss Kappa = " + round(fleiss_kappa) + "\n");

			System.out.println("Assumption for Cohens Kappa: There are exactly 2 annotators.");
			System.out.println("Cohens Kappa = " + round(cohen_kappa));
			System.out.println("Cohens Kappa (w.r.t. commercial) = " + round(cohen_c_kappa));
			System.out.println("Cohens Kappa (w.r.t. personal) = " + round(cohen_p_kappa));
			System.out.println("Cohens Kappa (w.r.t. informational) = " + round(cohen_i_kappa));
			System.out
					.println("\nNote: < 0	Poor agreement\n0.0 – 0.20	Slight agreement\n0.21 – 0.40	Fair agreement\n0.41 – 0.60	Moderate agreement\n0.61 – 0.80	Substantial agreement\n0.81 – 1.00	Almost perfect agreement");

		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	// ==========================================================================
	// PRINTING
	// ==========================================================================

	/**
	 * Print counted numbers.
	 * 
	 * @param title The title of the numbers to be printed
	 * @param map Stores the numbers to be printed.
	 */
	private static void printStatistics(String title, Map<String, Integer> map) {
		// Compute total value
		String sum = "Insgesamt";
		int total = 0;
		for (int number : map.values()) {
			total += number;
		}

		if (total > 0) {
			// Title
			System.out.println("\n");
			System.out.println(title + getSpace(title) + "\t" + "#");
			System.out.println("==============================================");
			// Print names and numbers
			for (String key : map.keySet()) {
				System.out.print(key + getSpace(key) + "\t");
				System.out.print(map.get(key).intValue() + "\t");
				System.out.println(round(map.get(key).intValue() / (double) total) + "%");
			}
			// Print sum
			System.out.println("----------------------------------------------");
			System.out.println(sum + getSpace(sum) + "\t" + total);
		}
	}

	// ==========================================================================
	// HELPERS
	// ==========================================================================

	/**
	 * Computes length of white space to be filled in after the given word
	 * 
	 * @param word The word
	 * @return The white space as a string
	 */
	private static String getSpace(String word) {
		String space = "";
		for (int i = 1; i <= 25 - word.length(); i++)
			space = space + " ";
		return space;
	}

	/**
	 * Round number to two decimal places
	 * 
	 * @param number The number to be rounded
	 * @return The rounded number
	 */
	private static double round(double number) {
		return Math.round(number * 10000) / 10000.0;
	}

}
