/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package de.upb.infexba.application;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.uima.UIMAFramework;
import org.apache.uima.analysis_engine.AnalysisEngine;
import org.apache.uima.cas.CAS;
import org.apache.uima.cas.FSIterator;
import org.apache.uima.cas.Feature;
import org.apache.uima.cas.FeatureStructure;
import org.apache.uima.cas.Type;
import org.apache.uima.cas.text.AnnotationFS;
import org.apache.uima.collection.CollectionReader;
import org.apache.uima.resource.ResourceSpecifier;
import org.apache.uima.util.XMLInputSource;

import de.upb.infexba.cr.XmiCollectionReader;

/**
 * Counts various numbers concerning the annotated classes of the input documents, i.e.,
 * the relevance, genre and the sentiment of the documents. Prints these numbers to
 * screen.
 * 
 * <p>
 * The application may take two arguments:
 * <ol type="1">
 * <li>The input directory containing the files to be processed. If you don't specify
 *     this directory, corpora/sentiment/smartphones is used.</li>
 * <li>The path to an XML descriptor for the type system to be used. If you don't specify
 *     this path, descriptors/infexbaTypeSystem.xml is used.</li>
 * </ol>
 * 
 * @author hwachsmuth
 *
 */
public class SentimentCompleteStatistics {
	
	// ==========================================================================
	// PATHS
	// ==========================================================================
	
	/**
	 * Location of collection reader
	 */
	private static String collectionReaderDesc = 
		"descriptors/collection/SentimentCorpusReader.xml";
	
	/**
	 * Location of type system
	 */
	private static String typeSystemDesc = 
		"descriptors/infexbaTypeSystem.xml";
	
	/**
	 * Location of analysis engine
	 */
	private static String analysisEngineDesc = 
		"descriptors/collection/DummyAnalysisEngine.xml";
	
	/**
	 * Path of input directory
	 */
	private static String inputDir = "corpora/lfa-corpus/corpus/smartphone_split/training/";

	
	
	// ==========================================================================
	//COUNTERS
	// ==========================================================================
	

	/**
	 * Map that stores the number of occurances of each of the four genre classes
	 * in the given corpus.
	 */
	private static Map<String, Integer> genreclassMap;

	/**
	 * Map that stores the number of occurances of each of the three polarities
	 * in the given corpus.
	 */
	private static Map<String, Integer> polarityMap;
	
	/**
	 * Map that stores the number of occurances of documents in the given corpus
	 * where relevance tag is true and where it is false, repesctively.
	 * 
	 */
	private static Map<String, Integer> relevanceMap;
	
	/**
	 * Map that stores the number of occurances of documents in the given corpus
	 * that contain 0, 1 or 2 metadata tags, respectively.
	 * 
	 */
	private static Map<String, Integer> metadataMap;
	
	
	// ==========================================================================
	// TYPE NAMES
	// ==========================================================================
	
	/**
	 * Name of the type of entities
	 */
	private static final String GENRE = "de.upb.infexba.Genre";

	/**
	 * Name of the type of entities
	 */
	private static final String RELEVANCE = "de.upb.infexba.Relevance";

	/**
	 * Name of the entity slot of a turnover statement
	 */
	private static final String OPINION = "de.upb.infexba.Opinion";
	
	/**
	 * Name of the type of declarations
	 */
	private static final String METADATA = "de.upb.infexba.Metadata";

	
	
	// ==========================================================================
	// COUNTING
	// ==========================================================================
	
	/**
	 * Main method iterates over all documents in the collection and sums up
	 * the number of included annotations.
	 * 
	 * @param args See class description
	 */
	public static void main(String[] args) {	
		// Set arguments if existing
		if (args.length > 0) inputDir = args[0];
		if (args.length > 1) typeSystemDesc = args[1]; 
		
		try {			
			// Get type system for initializing the counter map
			XMLInputSource in = new XMLInputSource(typeSystemDesc);
			
			// Initialize entries of maps for counting  		
			genreclassMap = new HashMap<String, Integer>();
			polarityMap = new HashMap<String, Integer>();
			relevanceMap = new HashMap<String, Integer>();	
			metadataMap = new HashMap<String, Integer>();	
			
			// Get the collection to be iterated over
			in = new XMLInputSource(collectionReaderDesc);
			ResourceSpecifier specifier = UIMAFramework.getXMLParser().parseResourceSpecifier(in);
			CollectionReader cr = UIMAFramework.produceCollectionReader(specifier);
			if (!inputDir.equals("")){
				cr.setConfigParameterValue(XmiCollectionReader.PARAM_INPUTDIR, inputDir);
				cr.reconfigure();
			}
			
			// Create an analysis engine, so we can create a CAS
			in = new XMLInputSource(analysisEngineDesc);
			specifier = UIMAFramework.getXMLParser().parseResourceSpecifier(in);
			AnalysisEngine ae = UIMAFramework.produceAnalysisEngine(specifier);
	  		CAS aCAS = ae.newCAS();	

			
			// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
			// Counting begins here...
					
	  		// Iterate over the whole collection
	  		System.out.print("Read documents...");
	  		int doc = 1;
			while (cr.hasNext()){
				cr.getNext(aCAS);
				int metadata = 0;
				// Iterate over all annotations of the document
			    FSIterator<AnnotationFS> fsIter = aCAS.getAnnotationIndex().iterator();
			    while (fsIter.hasNext()) {	    	
			    	// Get type of annotation
			    	Type type = fsIter.get().getType();
			    	
			    	if (type.getName().equals(METADATA)){
			    		metadata++;
			    	} else if (type.getName().equals(GENRE)){
			    		FeatureStructure fs = fsIter.get();
			    		List<Feature> features = fs.getType().getFeatures();
			    		Feature feature = features.get(3);
			    		String className = fs.getFeatureValueAsString(feature);
			    		if (genreclassMap.containsKey(className)){
			    			genreclassMap.put(className, genreclassMap.get(className)+1);
			    		} else genreclassMap.put(className, 1);
			    	}
			    	else if (type.getName().equals(OPINION)){
			    		FeatureStructure fs = fsIter.get();
			    		List<Feature> features = fs.getType().getFeatures();
			    		Feature feature = features.get(3);
			    		String className = fs.getFeatureValueAsString(feature);
			    		if (polarityMap.containsKey(className)){
			    			polarityMap.put(className, polarityMap.get(className)+1);
			    		} else polarityMap.put(className, 1);
			    	}
			    	else if (type.getName().equals(RELEVANCE)){
			    		FeatureStructure fs = fsIter.get();
			    		List<Feature> features = fs.getType().getFeatures();
			    		Feature feature = features.get(3);
			    		String className = fs.getFeatureValueAsString(feature);
			    		if (relevanceMap.containsKey(className)){
			    			relevanceMap.put(className, relevanceMap.get(className)+1);
			    		} else relevanceMap.put(className, 1);
			    	}
			    	
			    	

			    	fsIter.next();
			    }
			    
			    if (metadataMap.containsKey(""+metadata))
			    	metadataMap.put(""+metadata, metadataMap.get(""+metadata)+1);
			    else metadataMap.put(""+metadata, 1);
			    
		    	doc++;
		    	if (doc % 20 == 0) System.out.print(".");
			}

			// Counting ends here
			// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	
			// Print computed numbers
			printStatistics(METADATA, metadataMap);
			printStatistics(GENRE, genreclassMap);
			printStatistics(RELEVANCE, relevanceMap);
			printStatistics(OPINION, polarityMap);

		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	
	
	
	// ==========================================================================
	// PRINTING
	// ==========================================================================
	
	/**
	 * Print counted numbers.
	 * 
	 * @param title The title of the numbers to be printed
	 * @param map Stores the numbers to be printed.
	 */
	private static void printStatistics(String title, Map<String,Integer> map){
		// Compute total value
		String sum = "Insgesamt";
		int total = 0;
		for (int number : map.values()) {
			total += number;
		}
		
		if (total > 0){		
			// Title
			System.out.println("\n");
			System.out.println(title + getSpace(title) + "\t" + "#");
			System.out.println("==============================================");	
			// Print names and numbers
			for (String key : map.keySet()) {
				System.out.print(key + getSpace(key) + "\t");
				System.out.print(map.get(key).intValue() + "\t");
				System.out.println(round(map.get(key).intValue()/(double)total) + "%");
			}
			// Print sum
			System.out.println("----------------------------------------------");
			System.out.println(sum + getSpace(sum) + "\t" + total);		
		}
	}
	
	
	// ==========================================================================
	// HELPERS
	// ==========================================================================
	
	/**
	 * Computes length of white space to be filled in after the given word
	 * @param word The word
	 * @return The white space as a string
	 */
	private static String getSpace(String word){
		String space = "";
		for (int i=1; i<= 25-word.length(); i++) space = space + " ";
		return space;
	}
	
	/**
	 * Round number to two decimal places
	 * @param number The number to be rounded
	 * @return The rounded number
	 */
	private static double round(double number){
		return Math.round(number * 10000)/100.0;
	}
	
}
