package de.upb.infexba.linguistic;


import de.upb.infexba.util.WordLists;


/**
 * <p>
 * Class that stores several regular expressions in the context of
 * monetary expressions. These refer to the following categories:
 * <ul>
 *   <li>monetary subexpressions</li>
 *   <li>money terms</li>
 *   <li>units i.e. currencies and percentage</li>
 *   <li>fraction expressions</li>
 *   <li>number expressions</li>
 *   <li>prepositions and connectors</li>
 *   <li>fuzzy terms</li> 
 * </ul>
 * 
 * The common use is just to match with the pattern for monetary 
 * subexpressions since this spans all the different parts
 * defined by the categories given above.
 * 
 * TODO Recognition of expressions for the first character of a
 *   document. Generally, boundary recognition can be enhanced.
 * 
 * @author hwachsmuth
 *
 */
public class RegexMoney {
		
	// -------------------------------------------------------------------
	// IN BETWEEN
	// -------------------------------------------------------------------
	
	/**
	 * Arbitrarlily many blanks followed by a maximum of one line break
	 * ([CR], [LF] or [CR][LF], respectively) followed by arbitrarlily many 
	 * blanks.
	 */
	private static final String _ = "(\\p{Blank}*(\\r(\\n)?|\\n)?\\p{Blank}*|-)";
	
	/**
	 * Arbitrarlily many blanks followed by a maximum of one line break
	 * ([CR], [LF] or [CR][LF], respectively) followed by arbitrarlily many 
	 * blanks. However, at least one space character in total.
	 */
	private static final String __ = 
		"(" +
			"\\p{Blank}+(\\r(\\n)?|\\n)?" +
			"|(\\r(\\n)?|\\n)" +
		")" +
		"\\p{Blank}*";
	
	/**
	 * Filler words. A filler word is a real lower-case word here in that 
	 * it only consists of lower-case letters. 
	 */
	private static final String FILLER = 
		"[[a-z]äöüß]+";
	
	
	private static final String BRACKETED =
		"\\([^\\)]{1,40}\\)";
	
	
	
	// -------------------------------------------------------------------
	// INDICATOR WORDS
	// -------------------------------------------------------------------
	
	/**
	 * Possible leading prepositions of a monetary subexpression.
	 */
	private static final String PREP = 
		WordLists.loadWordList(WordLists.MONEY_PREPOSITIONS);
	
	/**
	 * Possible connectors between two monetary subexpressions.
	 */
	private static final String MONEY_CONNECTOR = 
		WordLists.loadWordList(WordLists.MONEY_CONNECTORS);

	/**
	 * Possible connectors between two number terms.
	 */
	private static final String CONNECTOR = 
		WordLists.loadWordList(WordLists.CONNECTORS);
	
	/**
	 * Words, which indicate that a number or a statement is fuzzy.
	 */
	private static final String FUZZY_TERM = 
		WordLists.loadWordList(WordLists.FUZZY_TERMS);
	
	
	
	// -------------------------------------------------------------------
	// DIGITS AND NUMBERS
	// -------------------------------------------------------------------

	/**
	 * Comma and decimal places
	 */
	private static final String DECIMAL_PLACES = 
		"((,|.)[0-9]+)";
	
	/**
	 * A natural number
	 */
	private static final String NATURAL = 
		"(([1-9][0-9]?(\\.[0-9]{3})+)" +
		"|(0|[1-9][0-9]*))";
	
	/**
	 * Character sequence that consists of a numeric value, i.e. a
	 * natural number with possible comma and decimal places
	 */
	private static final String NUMERIC_VALUE = 
		"(" + NATURAL + DECIMAL_PLACES + "?)";


	
	/**
	 * Words that name the first natural numbers. Supported numbers are
	 * given by the corresponding wordlist.
	 */
	private static final String SMALL_QUANTITY_WORD = 
		WordLists.loadWordList(WordLists.SMALL_QUANTITY_WORDS);
	
	/**
	 * Words that name big numbers such as millions or thousands and
	 * corresponding acronyms.
	 */
	private static final String BIG_QUANTITY_WORD = 
		WordLists.loadWordList(WordLists.BIG_QUANTITY_WORDS);

	/**
	 * Symbolic representation of signs of numbers.
	 */
	private static final String SIGN = "((\\+|–)" + _ + "|-)";
	
	/**
	 * Words that name signs of numbers.
	 */
	private static final String SIGN_WORD = 
		WordLists.loadWordList(WordLists.SIGN_WORDS);

	/**
	 * Complete term to specify a number. May consist of combinations of
	 * signs, numeric values and quantity words.
	 */
	private static final String NUMBER_TERM =
		"(" +
			"(" + 
				"((" + SIGN + "|" + SIGN_WORD + __ + ")?"+ NUMERIC_VALUE + ")" + 
				"|((" + SIGN_WORD + __ + ")?" + SMALL_QUANTITY_WORD + ")" + 
			")" +
			"((" + __ + BRACKETED + ")?" + __ + BIG_QUANTITY_WORD + ")?" +
		")";
	
	
	
	/**
	 * Complete expressions on numbers. May contain a fuzzy word, delare a 
	 * span, and so on.
	 */
	protected static final String NUMBER_EXPRESSION =
		"(" +
			"(" + FUZZY_TERM + __ + ")?" + 
			NUMBER_TERM + 
			"(" + 
				"(" + __ + CONNECTOR + __ + "|" + _ + "-" + _ + ")" + 
				"(" + FUZZY_TERM + __ + ")?" +
				NUMBER_TERM + 
			")?" +
		")";
	
	
	
	// -------------------------------------------------------------------
	// FRACTION EXPRESSIONS
	// -------------------------------------------------------------------
	
	/**
	 * Words that mean multiples. Supported fractions are declined 
	 * versions of the words given by the corresponding wordlist.
	 */
	protected static final String FRACTION_MULTIPLE =
		"((ver)?" + 
		WordLists.loadWordList(WordLists.MULTIPLE_FRACTIONS) +
		"((t(e)?|e|n)[mnrs]?)?)";
	
	/**
	 * Words that mean shares. Supported fractions are declined 
	 * versions of the words given by the corresponding wordlist.
	 */
	protected static final String FRACTION_SHARE = 
		WordLists.loadWordList(WordLists.SHARE_FRACTIONS);
	
	/**
	 * Complete fraction expressions which may contain
	 * a fuzzy word and a defining article, and must contain
	 * a multiple or a share word.
	 */
	private static final String FRACTION_EXPRESSION =
		"(" +
			"(" + 
				FUZZY_TERM + __ + 
				"(" + WordLists.loadWordList(WordLists.DETERMINERS) + __ + ")?" +
			")?" + 
			"(" + 
				FRACTION_MULTIPLE + 
				"|(" + SMALL_QUANTITY_WORD + __ + ")?" + FRACTION_SHARE +
			")" + 
		")";
	
	
	
	// -------------------------------------------------------------------
	// UNITS
	// -------------------------------------------------------------------
	
	/**
	 * Words and symbols that mean percentage and the like.
	 */
	private static final String PERCENTAGE = 
		WordLists.loadWordList(WordLists.PERCENTAGES);
		
	/**
	 * Terms and symbols that refer to any currency. Supported currencies
	 * are given by the corresponding word list.
	 */
	private static String CURRENCY = 
		WordLists.loadWordList(WordLists.CURRENCIES);
	
	
	
	// -------------------------------------------------------------------
	// MONEY TERMS
	// -------------------------------------------------------------------
	
	/**
	 * Relative terms on money, Such a term consist of a number
	 * expression followed by a percentage term or of a fraction expression.
	 *
	 */
	public static final String RELATIVE_MONEY_TERM =
		"(" + NUMBER_EXPRESSION + _ + PERCENTAGE + 
		"|" + FRACTION_EXPRESSION + ")";
	
	/**
	 * Absolute terms on money. Such a term consists of a number expression 
	 * followed by a currency expression or vice versa.
	 *
	 */
	public static final String ABSOLUTE_MONEY_TERM =
		"(" + 
			"("+ NUMBER_EXPRESSION + "(" + __ + BRACKETED +")?" + _ + CURRENCY + 
			"|" + CURRENCY + _ + NUMBER_EXPRESSION + ")" +
			"(" + __ + "\\(" + NUMBER_EXPRESSION + _ + PERCENTAGE + "\\)"+ ")?" +
		")";
	

	
	/**
	 * Absolute or relative money terms.
	 */
	private static final String MONEY_TERM =
		"(" + ABSOLUTE_MONEY_TERM + "|" + RELATIVE_MONEY_TERM + ")";
	
	/**
	 * Regular expression for absolute or relative terms on money possibly
	 * preceded by a preposition and a filler word.
	 * <p>
	 * <b>NOTICE:</b> It is assumed that monetary subexpressions start and
	 * end with a non-letter character to ensure that there is a word boundary 
	 * at the beginning and the end of it. That is, only take the substring 
	 * starting at index 1 and ending at index length-1 if you want to have 
	 * the actual subexpression.
	 */
	public static final String MONEY_SUBEXPRESSION =
		"(([^a-zA-Z_0-9_äöüß])(" +
			"(" + PREP + __ + "("+ FILLER + __ + ")?" + ")?" + 
			MONEY_TERM +
			"(" + __ + MONEY_CONNECTOR + __ + MONEY_TERM + ")?" +
		")([^a-zA-Z_0-9_äöüß]))";
}
