package de.upb.infexba.linguistic;

import de.upb.infexba.util.WordLists;

/**
 * <p>
 * Class that stores several regular expressions in the context of
 * time expressions. These refer to the following categories:
 * <ul>
 *   <li>time expressions, themselves</li>
 *   <li>date expressions</li>
 *   <li>period expressions</li>
 *   <li>attributes, i.e., numbers or relational terms</li>
 *   <li>prepositions, determiners and connectors</li>
 * </ul>
 * 
 * The common use is just to match with the pattern for time 
 * subexpressions since this spans all the different parts
 * defined by the categories given above.
 * 
 * TODO Avoid the leading preposition "zum" if it is preceded by expressions
 *   like "im Vergleich" or the like.
 * 
 * @author hwachsmuth
 *
 */
public class RegexTime {

	// -------------------------------------------------------------------
	// IN BETWEEN
	// -------------------------------------------------------------------
	
	/**
	 * Arbitrarlily many blanks followed by a maximum of one line break
	 * ([CR], [LF] or [CR][LF], respectively) followed by arbitrarlily many 
	 * blanks. However, at least one space character in total.
	 */
	private static final String __ = 
		"(" + "\\p{Blank}+(\\r(\\n)?|\\n)?" + "|(\\r(\\n)?|\\n)" + ")" + "\\p{Blank}*";

	/**
	 * Filler words. A filler word is a real lower-case word here in that 
	 * it only consists of lower-case letters. 
	 */
	protected static final String FILLER = 
		"[[a-z]äöüß]+";
	
	
	
	// -------------------------------------------------------------------
	// BASIC TERMS
	// -------------------------------------------------------------------
	
	/**
	 * Determiners.
	 */
	protected static final String DET = 
		WordLists.loadWordList(WordLists.DETERMINERS);
	
	/**
	 * Words that connect two time expressions.
	 */
	protected static final String CONNECTOR =
		WordLists.loadWordList(WordLists.CONNECTORS);

	/**
	 * Possible leading prepositions of a time expression.
	 */
	protected static final String PREP =
		WordLists.loadWordList(WordLists.TIME_PREPOSITIONS);
	
	/**
	 * Terms consisting of a preposition and an optional determiner.
	 */
	protected static final String PREP_TERM =
		"("+ PREP + "("+ __ + DET + ")?)";
	

	
	// -------------------------------------------------------------------
	// ATTRIBUTES
	// -------------------------------------------------------------------
	
	/**
	 * Numbers between 1 and 99 or words to name a small quantity as
	 * defined by the corresponding word list.
	 */
	protected static final String QUANTITY =
		"((1|2|3|4|5|6|7|8|9)\\d?" +
		"|" + WordLists.loadWordList(WordLists.SMALL_QUANTITY_WORDS) +")";
	
	/**
	 * Ordinal numbers between 1st and and 12th or words to name an
	 * ordinal number as defined by the corresponding word list.
	 */
	protected static final String ORDINAL =
		"((1[012]?|2|3|4|5|6|7|8|9)(\\.|te[mnrs]?)" +
		"|" + WordLists.loadWordList(WordLists.ORDINAL_WORDS) +"([mnrs])?)";
	
	/**
	 * Terms that specify a quantity or an ordinal number.
	 */
	protected static final String NUMERIC_TERM =
		"(" + 
			QUANTITY + 
			"|" + ORDINAL + "(-" + ORDINAL + ")?(" + __ + QUANTITY + ")?" + ")";
	
	/**
	 * Terms that specify a time relation such as "next", "following"
	 * or "current".
	 */
	protected static final String RELATIONAL_TERM = 
		"(" + WordLists.loadWordList(WordLists.DIRECTIONAL_TERMS) + "[mnrs]?)";


	
	// -------------------------------------------------------------------
	// PERIODS
	// -------------------------------------------------------------------
	
	/**
	 * Words to specify the boundary of a period.
	 */
	protected static final String BOUNDARY =
		WordLists.loadWordList(WordLists.BOUNDARY_WORDS);

	/**
	 * Terms consisting of a boundary word and an optional determiner.
	 */
	protected static final String BOUNDARY_TERM =
		"("+ BOUNDARY + "(" + __ + DET + ")?" + ")";
	
	/**
	 * Terms for regular periods of time like years or quarters. 
	 */
	public static final String PERIOD_REGULAR =
		"(" +
			"Q(1|2|3|4)|H(1|2)(\\/(19|20)?\\d{2})?" +
			"|(" +
				"(" + "([Hh]albe(n|s)?" + __ + ")?(\\w([a-z[äöüß]])*((-J|j)ahr|(-Q|q)uartal)|[Jj]ahr|[Qq]uartal))" +
				"(s|e(n|s)?)?" +
				"([a-z[äöüß]])*" +
			")" +
		")";
	
	/**
	 * Terms for months as defined by the corresponding word list.
	 */
	public static final String PERIOD_MONTH =
		"(((Monat(e(n|s)?|s)?|Zeitraum)" + __ + "(von" + __ +")?)?" + WordLists.loadWordList(WordLists.MONTHS) + "|Monat(e(n|s)?|s)?([a-z[äöüß]])*)";
	
	/**
	 * Terms to specify relative period of time like "reported time" as 
	 * defined by the corresponding word list.
	 */
	public static final String PERIOD_RELATIVE =
		"("+ WordLists.loadWordList(WordLists.RELATIVE_PERIODS)+"(s|es)?)";
	
	/**
	 * Any terms to specify a period of time.
	 */
	public static final String PERIOD_TERM =
		"(" + PERIOD_REGULAR + "|" + PERIOD_MONTH + "|" + PERIOD_RELATIVE + ")";
	
	/**
	 * Numeric terms to specify a year.
	 */
	public static final String YEAR_TERM =
		"((19|20)\\d{2}(/(19|20)?\\d{2})?" + "|\\d{2}/\\d{2})";
	
	/**
	 * Terms to specify a period of time or a certain year.
	 */
	protected static final String PERIOD =
		"(" + PERIOD_TERM + "("+ __ + YEAR_TERM + ")?|"+ YEAR_TERM + ")";

	/**
	 * This is a fix for the words Viertel and Haelfte, which are also often
	 * used in MonetaryExpressions. A more accurate modeling of corresponding
	 * time expressions would require to revise the whole regex. However, this
	 * here should work just fine.
	 */
	protected static final String DIRTY_FIX =
		"(" +
			"(" +
				"(" + RELATIONAL_TERM + "|" + ORDINAL + ")" + 
				"(s|e(n|s)?)?" +
			")" +
			"Viertel|Hälfte|Haelfte" +
		")";
	
	/**
	 * Expressions that consist of a period term preceded by numeric and/or 
	 * relational attributes.
	 */
	protected static final String PERIOD_SUBEXPRESSION =
		"(" +
			"(" +
				"(" + RELATIONAL_TERM + "(" + __ + NUMERIC_TERM + ")?" + __ + ")" + 
				"|(" + NUMERIC_TERM + "(" + __ + RELATIONAL_TERM + ")?" + __ + ")" +
			")?" +
			PERIOD +
		")";
	
	/**
	 * Expressions that consist of one or more connected period subexpressions
	 * possibly preceded by a prepositional term and/or a boundary term, and,
	 * if so, a filler word in between them.
	 */
	public static final String PERIOD_EXPRESSION =
		"(" +
			"(" + 
				"(" +
					PREP_TERM + __ + "(" + BOUNDARY_TERM + __ + ")?" +
					"|" + BOUNDARY_TERM + __ +
				")" +
				"(" + FILLER + __ + ")?" +
			")?" + 
			"(" + PERIOD_SUBEXPRESSION + "|" + DIRTY_FIX + ")" +
			"(" + 
				__ + CONNECTOR +
				"(" + __ + BOUNDARY_TERM + ")?" +
				"(" + __ + FILLER + ")?" +
				__ + PERIOD_SUBEXPRESSION + 
			")*" +
		")";
	
	

	// -------------------------------------------------------------------
	// DATES
	// -------------------------------------------------------------------
	
	/**
	 * Numeric representations of days in a date.
	 */
	protected static final String DATE_DAY =
		"((0?[123456789]|[12]\\d|3[01])(\\.|/))";
	
	/**
	 * Numeric representations of months in a date.
	 */
	protected static final String DATE_MONTH =
		"((0?[123456789]|1[012])(\\.|/))";
	
	/**
	 * Numeric representations of years in a date.
	 */
	protected static final String DATE_YEAR =
		"((19|20)?\\d{2})";
	
	/**
	 * Representations of dates as given by combinations of days, months
	 * and years possibly preceded by a prepositional term. Note that
	 * months may be specified in a numerical or a textual way.
	 */
	public static final String DATE_EXPRESSION =
		"(" + 
			"(" + PREP_TERM + __ + ")?" +
			"(" + DATE_DAY + "(" + __ + ")?)" +
			"(" + PERIOD_MONTH + "|" + DATE_MONTH + ")" +
			"(" + "( )?" + DATE_YEAR + ")?" +
		")";
	

	
	// -------------------------------------------------------------------
	// TIME EXPRESSIONS
	// -------------------------------------------------------------------
	
	/**
	 * Expressions that specify a date or a period of time.
	 */
	public static final String TIME_EXPRESSION =
		"\\b(" +
			DATE_EXPRESSION + "|" + PERIOD_EXPRESSION +
		")";
		
	/**
	 * Expressions that specify a date.
	 */
	public static final String DATE =
		"\\b(" +
			DATE_EXPRESSION +
		")\\b";
	
	/**
	 * Expressions that specify a period of time.
	 */
	public static final String PERIOD_TIME_EXPRESSION =
		"\\b(" +
			PERIOD_EXPRESSION +
		")\\b";
		
		
		
		
		
		
		
		
		
		
		
		
		
		

}
