package de.upb.infexba.ml.feature;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.TreeSet;

import de.upb.infexba.ml.IntentionFeatureCalculator;
import de.upb.infexba.util.Tuple;

/**
 * The GenreFeatureHandler adds further functionality to the FeatureHandler such as the calculation of genre-specific
 * features and the top-features per genre-class.
 * 
 * @author kabu
 * 
 */
abstract public class GenreFeatureHandler extends FeatureHandler {

	protected static final byte GENRE_SPECIFIC_FEATURE = 2;

	protected static final byte TOP_FEATURE = 3;

	private static final String GENRE_SPECIFIC_SUFFIX = "|gs";

	private static final String TOP_SUFFIX = "|top";

	public static final byte UNDEFINED_KIND = -1;

	// Note:
	// GS-TOP-suffix will be added during computation of the features and
	// will be removed here before it's stored as a feature

	protected static String addGsTopSuffix(String key, short kind) {
		if (kind == GENRE_SPECIFIC_FEATURE)
			return key + GENRE_SPECIFIC_SUFFIX;
		else if (kind == TOP_FEATURE)
			return key + TOP_SUFFIX;
		else
			return key;
	}

	protected static String removeGsTopSuffix(String feature_name) {
		if (feature_name.indexOf(GENRE_SPECIFIC_SUFFIX) > -1)
			return feature_name.substring(0, feature_name.indexOf(GENRE_SPECIFIC_SUFFIX));
		else if (feature_name.indexOf(TOP_SUFFIX) > -1)
			return feature_name.substring(0, feature_name.indexOf(TOP_SUFFIX));
		else
			return feature_name;
	}
	
	public static boolean isGenreSpecificFeature(String feature_name) {
		return feature_name.endsWith(GENRE_SPECIFIC_SUFFIX);
	}

	public static boolean isTopFeature(String feature_name) {
		return feature_name.endsWith(TOP_SUFFIX);
	}

	public void addFeature(String feature) {
		String key = removeGsTopSuffix(feature);
		super.addFeature(key);
	}

	public void setFeatures(String[] features) {
		String key;
		for (String feature : features) {
			key = removeGsTopSuffix(feature);
			super.addFeature(key);
		}
	}

	public boolean handles(String feature) {
		return super.handles(removeGsTopSuffix(feature));
	}

	public double getFromFeatureMap(String feature) {
		String key = removeGsTopSuffix(feature);
		return super.getFromFeatureMap(key);
	}

	/**
	 * returns the map that (internally, temporarily) stores information w.r.t a certain classification. The Parameter
	 * might be useful to distinguish between different maps, e.g. see lemma and word maps in VocabularyFeatureHandler.
	 * 
	 * @param name
	 * @param kind
	 * @return
	 */
	abstract protected HashMap<String, Double> getMapByName(String classification, short kind);

	protected HashMap<String, Double> getMapByName(String classification) {
		return getMapByName(classification, UNDEFINED_KIND);
	}

	/**
	 * Assume you've given one of the possible genre classes and now you want to know the other two classes, then you
	 * can use this method.
	 * 
	 * @param classification
	 * @return
	 */
	private static String getSecondClass(String classification) {
		if (classification.equals(IntentionFeatureCalculator.PERSONAL))
			return IntentionFeatureCalculator.INFORMATIONAL;
		if (classification.equals(IntentionFeatureCalculator.COMMERCIAL))
			return IntentionFeatureCalculator.INFORMATIONAL;
		// if INFORMATIONAL
		return IntentionFeatureCalculator.PERSONAL;
	}
	
	/**
	 * Assume you've given one of the possible genre classes and now you want to know the other two classes, then you
	 * can use this method.
	 * 
	 * @param classification
	 * @return
	 */
	private static String getThirdClass(String classification) {
		if (classification.equals(IntentionFeatureCalculator.PERSONAL))
			return IntentionFeatureCalculator.COMMERCIAL;
		if (classification.equals(IntentionFeatureCalculator.COMMERCIAL))
			return IntentionFeatureCalculator.PERSONAL;
		// if INFORMATIONAL
		return IntentionFeatureCalculator.COMMERCIAL;
	}

	// -------------------------------------------------------------------
	// TOP-WORDS
	// -------------------------------------------------------------------

	public TreeSet<String> getTop(int t, int skip_first_x_words) {
		String name1 = IntentionFeatureCalculator.PERSONAL;
		String name2 = IntentionFeatureCalculator.COMMERCIAL;
		String name3 = IntentionFeatureCalculator.INFORMATIONAL;
		return getTop(t, name1, name2, name3, UNDEFINED_KIND, skip_first_x_words);
	}

	public TreeSet<String> getTop(int t, short kind, int skip_first_x_words) {
		String name1 = IntentionFeatureCalculator.PERSONAL;
		String name2 = IntentionFeatureCalculator.COMMERCIAL;
		String name3 = IntentionFeatureCalculator.INFORMATIONAL;
		return getTop(t, name1, name2, name3, kind, skip_first_x_words);
	}

	/**
	 * Get the T words that appear most frequently in only one map.
	 * 
	 * @param T
	 * @param map
	 * @return
	 */
	protected TreeSet<String> getTop(int T, String name1, String name2, String name3, short kind, int skip_first_x_words) {
		Set<String> all_top_words = new HashSet<String>();
		List<String> top_personal_words = this.getTop(T, name1, kind);
		List<String> top_commercial_words = this.getTop(T, name2, kind);
		List<String> top_informational_words = this.getTop(T, name3, kind);

		// merge
		if (skip_first_x_words == 0) {
			all_top_words.addAll(top_personal_words);
			for (String word : top_commercial_words)
				if (!all_top_words.contains(word))
					all_top_words.add(word);
			for (String word : top_informational_words)
				if (!all_top_words.contains(word))
					all_top_words.add(word);
		} else {
			for (int i = 0; i < T && i < top_personal_words.size(); i++)
				if (i >= skip_first_x_words)
					all_top_words.add(top_personal_words.get(i));
			for (int i = 0; i < T && i < top_commercial_words.size(); i++)
				if (i >= skip_first_x_words)
					all_top_words.add(top_commercial_words.get(i));
			for (int i = 0; i < T && i < top_informational_words.size(); i++)
				if (i >= skip_first_x_words)
					all_top_words.add(top_informational_words.get(i));
		}

		// Mark Features as "top" features
		TreeSet<String> output = new TreeSet<String>();
		for (String feature : all_top_words)
			output.add(addGsTopSuffix(feature, TOP_FEATURE));

		return output;
	}

	/**
	 * Get the T words that appear most frequently in the map.
	 * 
	 * @param T
	 * @param map
	 * @return
	 */
	protected List<String> getTop(int T, String name, short kind) {
		HashMap<String, Double> map = getMapByName(name, kind);

		// create list of tuples (lemma, value)
		List<Tuple> list = new ArrayList<Tuple>();
		for (String key : map.keySet())
			list.add(new Tuple(key, map.get(key)));

		// sort
		Collections.sort(list);

		// copy to a string-array that contains the most frequent words
		List<String> top = new ArrayList<String>();
		for (int i = 0; (i < T) && (i < list.size()-1); i++) {
			top.add(list.get(i).getString());
		}

		print_top_words(top, name, map, kind);

		return top;
	}

	// -------------------------------------------------------------------
	// GENRE-SPECIFIC-WORDS
	// -------------------------------------------------------------------

	/**
	 * Get the T words that appear more frequently in only one of the maps...
	 */
	public TreeSet<String> getAllGenreSpecific(double threshold, int min, boolean normalize__sum) {
		return getAllGenreSpecific(-1, UNDEFINED_KIND, threshold, min, normalize__sum);
	}

	/**
	 * Get the T words that appear more frequently in only one of the maps...
	 */
	public TreeSet<String> getAllGenreSpecific(int t, double threshold, int min, boolean normalize__sum) {
		return getAllGenreSpecific(t, UNDEFINED_KIND, threshold, min, normalize__sum);
	}

	/**
	 * Get the T words that appear more frequently in only one of the maps...
	 */
	public TreeSet<String> getAllGenreSpecific(short kind, double threshold, int min, boolean normalize__sum) {
		return getAllGenreSpecific(-1, kind, threshold, min, normalize__sum);
	}

	/**
	 * Get the T words that appear more frequently in only one of the maps...
	 */
	public TreeSet<String> getAllGenreSpecific(int t, short kind, double threshold, int min, boolean normalize__sum) {

		String name1 = IntentionFeatureCalculator.PERSONAL;
		String name2 = IntentionFeatureCalculator.COMMERCIAL;
		String name3 = IntentionFeatureCalculator.INFORMATIONAL;

		List<String> specific_map1_words = this.getGenreSpecificForOneClass(t, kind, name1, threshold, min, normalize__sum);
		List<String> specific_map2_words = this.getGenreSpecificForOneClass(t, kind, name2, threshold, min, normalize__sum);
		List<String> specific_map3_words = this.getGenreSpecificForOneClass(t, kind, name3, threshold, min, normalize__sum);

		// merge
		TreeSet<String> all_specific_words = new TreeSet<String>();
		all_specific_words.addAll(specific_map1_words);
		for (String word : specific_map2_words)
			if (!all_specific_words.contains(word))
				all_specific_words.add(word);
		for (String word : specific_map3_words)
			if (!all_specific_words.contains(word))
				all_specific_words.add(word);
		return all_specific_words;
	}

	/**
	 * Get the T words that appear more frequently in map1, but not in map2 or map3.
	 */
	public List<String> getGenreSpecificForOneClass(int t, String name1, double threshold, int min, boolean normalize__sum) {
		return getGenreSpecificForOneClass(t, UNDEFINED_KIND, name1, threshold, min, normalize__sum);
	}

	/**
	 * Get the T words that appear more frequently in map1, but not in map2 or map3.
	 */
	public List<String> getGenreSpecificForOneClass(int t, short kind, String name1, double threshold, int min, boolean normalize__sum) {

		String name2 = getSecondClass(name1);
		String name3 = getThirdClass(name1);

		HashMap<String, Double> map1 = getMapByName(name1, kind);
		HashMap<String, Double> map2 = getMapByName(name2, kind);
		HashMap<String, Double> map3 = getMapByName(name3, kind);

		System.out.print(name1 + ":  ");

		List<String> genre_specific_words = new ArrayList<String>();

		List<Tuple> list12 = get_genre_specific_ratio(map1, name1, map2, name2, min, normalize__sum);
		Collections.sort(list12);
		List<Tuple> list13 = get_genre_specific_ratio(map1, name1, map3, name3, min, normalize__sum);
		Collections.sort(list13);

		// don't add words twice...
		Set<String> merge_set = new HashSet<String>();

		// add words as features
		int counter = 0;
		Iterator<Tuple> iter12 = list12.iterator();
		Iterator<Tuple> iter13 = list13.iterator();
		Tuple tuple;
		boolean iter12_exceeded_threshold = false, iter13_exceeded_threshold = false;
		while ((iter12.hasNext() || iter13.hasNext()) && !(iter12_exceeded_threshold && iter13_exceeded_threshold)) {
			if (iter12.hasNext() && !iter12_exceeded_threshold) {
				tuple = iter12.next();
				// System.err.println(counter +" = "+t.value+ "  (threshold = "+threshold+")");
				if (tuple.getValue() > threshold && !merge_set.contains(tuple.getString())) {
					genre_specific_words.add(tuple.getString());
					merge_set.add(tuple.getString());

					System.out.println(tuple.getString()); // + " (%=" + tuple.getValue() + ", abs=" + map1.get(tuple.getString()) + "), "); // TODO: 

					counter++;
					if (counter == t)
						break;
				} else
					iter12_exceeded_threshold = true;
			}
			if (iter13.hasNext() && !iter13_exceeded_threshold) {
				tuple = iter13.next();
				// System.err.println(counter +" = "+t.value+ "  (threshold = "+threshold+")");
				if (tuple.getValue() > threshold && !merge_set.contains(tuple.getString())) {
					genre_specific_words.add(tuple.getString());
					merge_set.add(tuple.getString());

					System.out.println(tuple.getString()); // + " (%=" + tuple.getValue() + ", abs=" + map1.get(tuple.getString()) + "), "); // TODO

					counter++;
					if (counter == t)
						break;
				} else
					iter13_exceeded_threshold = true;
			}
		}
		System.out.print("\n");

		// Mark features as "GenreSpecific" features
		List<String> output = new ArrayList<String>();
		for (String feature : genre_specific_words)
			output.add(addGsTopSuffix(feature, GENRE_SPECIFIC_FEATURE));

		return output;
	}

	/**
	 * Get the ratio (relative #occurrences of word w in map 1)/(relative #occurrences of word w in map 2) for all words
	 * w in map1.
	 * 
	 * @param T
	 * @param map1
	 * @param map2
	 * @return
	 */
	private static List<Tuple> get_genre_specific_ratio(HashMap<String, Double> map1, String name1, HashMap<String, Double> map2, String name2, int min,
			boolean normalize__sum) {
		Double sum1 = 1., sum2 = 1.;
		if (normalize__sum) {
			for (String key : map1.keySet())
				sum1 += map1.get(key);
			for (String key : map2.keySet())
				sum2 += map2.get(key);
		}

		// create sorted lists of tuples (lemma, #occurrences)
		List<Tuple> list = new ArrayList<Tuple>();
		for (String lemma : map1.keySet())
			list.add(new Tuple(lemma, map1.get(lemma)));
		Collections.sort(list);

		// list with all words and a ratio that denotes how specific they are for map1
		List<Tuple> specific_ratio = new ArrayList<Tuple>();
		for (int i = 0; i < list.size(); i++) {
			Tuple tuple = list.get(i);
			// if the word occurs at least min times...
			if (tuple.getValue() >= min)
				if (!map2.containsKey(tuple.getString()))
					// if it doesn't appear in map2 -> then ratio = relative #occurrences of the lemma in map1
					specific_ratio.add(new Tuple(tuple.getString(), tuple.getValue() / sum1));
				else
					// if it does store the ratio
					specific_ratio.add(new Tuple(tuple.getString(), (tuple.getValue() / sum1) / (map2.get(tuple.getString()) / sum2)));
		}
		Collections.sort(specific_ratio);

		return specific_ratio;
	}

	static void print_top_words(List<String> feature_names, String genre, HashMap<String, Double> map, int kind) {
		System.out.print(genre + ":  ");
		for (String feature_name : feature_names)
			System.out.print("\"" + feature_name + "\" (" + map.get(feature_name) + "), ");
		System.out.print("\n");
	}
}
