package de.upb.infexba.ml.feature.partofspeech;

import java.util.Arrays;
import java.util.TreeSet;

import de.upb.infexba.Token;
import de.upb.infexba.ml.feature.FeatureHandler;
import de.upb.infexba.ml.feature.vocabulary.VocabularyFeatureHandler;

public class PartOfSpeechFeatureHandler extends FeatureHandler {

	// -------------------------------------------------------------------
	// CONSTANTS FOR SPECIAL FEATURES
	// -------------------------------------------------------------------

	public static final String POS_SET_NOUN_OR_ADJECTIVE = "8_POS_SET_NOUN_OR_ADJECTIVE", POS_SET_PERSONAL_OR_POSSESSIV = "8_POS_SET_PERSONAL_OR_POSSESSIV";

	/**
	 * contains all single POS tags plus all tags given in PartOfSpeechConstants.SETS.
	 */
	public static final String[] ALL_AND_GROUPS = { /* 1 */"ADJA", "ADJD", "ADV", "APPR", "APPRART", "APPO", "APZR", "ART", "CARD", "FM", "ITJ", "KOUI",
			"KOUS", "KON", "KOKOM", "NN", "NE", /* 2 */
			"PDS", "PDAT", "PIS", "PIAT", "PIDAT", "PPER", "PPOSS", "PPOSAT", "PRELS", "PRELAT", "PRF", "PWS", "PWAV", "PAV", "PTKZU", "PTKNEG", "PTKVZ",
			"PTKANT", "PTKA", "TRUNC", /* 3 */"VVFIN", "VVIMP", "VVINF", "VVIZU", "VVPP", "VAFIN", "VAIMP", "VAINF", "VAPP", "VMFIN", "VMINF", "VMPP",/* 4 */
			PartOfSpeechConstants.XY, PartOfSpeechConstants.KOMMA, PartOfSpeechConstants.SATZENDE, PartOfSpeechConstants.SATZINTERN, /* Groups */
			POS_SET_NOUN_OR_ADJECTIVE, POS_SET_PERSONAL_OR_POSSESSIV };

	public static final String[] SETS = { POS_SET_NOUN_OR_ADJECTIVE, POS_SET_PERSONAL_OR_POSSESSIV };

	// all pos features that can be handled
	private static final TreeSet<String> all_pos_features = new TreeSet<String>(Arrays.asList(ALL_AND_GROUPS));

	// -------------------------------------------------------------------
	// What shall be counted?
	// -------------------------------------------------------------------

	public PartOfSpeechFeatureHandler() {
		super(ALL_AND_GROUPS);
	}

	public PartOfSpeechFeatureHandler(String[] pos) {
		super(pos);
	}

	public TreeSet<String> getFeatures() {
		if (this.handles_given_features)
			return this.features;
		else
			return all_pos_features;
	}

	public boolean canHandle(String feature) {
		return PartOfSpeechFeatureHandler.all_pos_features.contains(feature);
	}

	// -------------------------------------------------------------------
	// Counting.
	// -------------------------------------------------------------------

	private String getFeatureName(String pos) {
		return PartOfSpeechConstants.getPosAsText(pos);
	}

	public void handle(String pos) {
		String feature = this.getFeatureName(pos);
		this.increaseInFeatureMap(feature, 1.);

		// special handling of POS-sets
		// TODO: will always be added...
		if (is_in_list(PartOfSpeechConstants.NOUN_ADJ, pos))
			this.increaseInFeatureMap(POS_SET_NOUN_OR_ADJECTIVE, 1.);

		if (is_in_list(PartOfSpeechConstants.PERS_PROSSESSIV, pos))
			this.increaseInFeatureMap(POS_SET_PERSONAL_OR_POSSESSIV, 1.);
	}

	public static String getPos(Token token) {
		String pos = token.getPos();
		String text = token.getCoveredText();
		if (VocabularyFeatureHandler.XY_SYMBOLS.contains(text)) {
			pos = "XY";
		} else if (text.equals("€")) {
			pos = "NN";
		} else if (VocabularyFeatureHandler.QUOTE_LIST.contains(text)) {
			pos = PartOfSpeechConstants.SATZINTERN;
		} else if (VocabularyFeatureHandler.PARANTHESIS_LIST.contains(text)) {
			pos = PartOfSpeechConstants.SATZINTERN;
		}
		return pos;
	}

	private boolean is_in_list(String[] list, String s) {
		for (String t : list)
			if (s.equals(t))
				return true;
		return false;
	}

	public static final short MY_NORMALIZATION_TEST = -5; // that was just a test

	public double getNormalizedValue(String feature, double value, Double number_of_tokens) {
		if (this.normalizationMode == NO_FACTORS)
			return 1. * value / number_of_tokens;

		double factor = 1.;

		if (this.normalizationMode == MY_NORMALIZATION) {

			// old normalization
			if (feature.equals(PartOfSpeechFeatureHandler.POS_SET_NOUN_OR_ADJECTIVE))
				factor = 1.;
			else if (feature.equals("NN") || feature.equals(PartOfSpeechFeatureHandler.POS_SET_PERSONAL_OR_POSSESSIV))
				factor = 4.;
			else if (feature.equals("NE"))
				factor = 5.;
			else if (feature.equals("ART") || feature.equals("ADJA") || feature.equals("ADJD") || feature.equals("ADV") || feature.equals("APPR")
					|| feature.equals("VVFIN") || feature.equals("VAIMP") || feature.equals("KON"))
				factor = 10.;
			else if (feature.equals("VAFIN") || feature.equals("XY") || feature.equals("CARD"))
				factor = 20.;
			else if (feature.equals("VVINF") || feature.equals("APPRART") || feature.equals("VVPP"))
				factor = 30.;
			else if (feature.equals("KOUS") || feature.equals("KOKOM") || feature.equals("PPER") || feature.equals("KOMMA"))
				factor = 40.;
			else if (feature.equals("PTKNEG") || feature.equals("PTKVZ"))
				factor = 60.;
			else if (feature.equals("PDS") || feature.equals("SATZINTERN") || feature.equals("PDAT") || feature.equals("PIS") || feature.equals("PIAT")
					|| feature.equals("PPOSS") || feature.equals("PPOSAT") || feature.equals("PRELS") || feature.equals("PRF") || feature.equals("PTKZU")
					|| feature.equals("VVIZU") || feature.equals("VAINF") || feature.equals("SATZENDE") || feature.equals("FM"))
				factor = 100.;
			else if (feature.equals("VMFIN") || feature.equals("APZR") || feature.equals("KOUI") || feature.equals("PWAV"))
				factor = 200.;
			else if (feature.equals("PRELAT") || feature.equals("PWS"))
				factor = 300.;
			else if (feature.equals("VAPP") || feature.equals("ITJ") || feature.equals("PTKANT") || feature.equals("PTKA") || feature.equals("PRELAT")
					|| feature.equals("TRUNC"))
				factor = 400.;
			else if (feature.equals("VVIMP") || feature.equals("PAV"))
				factor = 600.;
			else {
				// System.err.println("ist POS, aber oben nicht erfasst: " + feature);
				factor = 1.;
			}
		} else if (this.normalizationMode == MY_NORMALIZATION_TEST) {

			// new "conservative" approximation/factors
			if (feature.equals(PartOfSpeechFeatureHandler.POS_SET_NOUN_OR_ADJECTIVE))
				factor = 1.;
			else if (feature.equals("NN") || feature.equals("XY"))
				factor = 3.;
			else if (feature.equals(PartOfSpeechFeatureHandler.POS_SET_PERSONAL_OR_POSSESSIV) || feature.equals("ADJA") || feature.equals("$(")
					|| feature.equals("$."))
				factor = 4.;
			else if (feature.equals("NE"))
				factor = 5.;
			else if (feature.equals("ART") || feature.equals("ADV") || feature.equals("FM"))
				factor = 6.;
			else if (feature.equals("$,") || feature.equals("PPER"))
				factor = 9.;
			else if (feature.equals("ADJD") || feature.equals("APPR") || feature.equals("VVFIN") || feature.equals("VAIMP") || feature.equals("KON")
					|| feature.equals("TRUNC") || feature.equals("VAFIN"))
				factor = 10.;
			else if (feature.equals("PIS") || feature.equals("PPOSAT") || feature.equals("VVINF") || feature.equals("PDAT"))
				factor = 15.;
			else if (feature.equals("PDS") || feature.equals("KOUS") || feature.equals("PIAT") || feature.equals("CARD"))
				factor = 20.;
			else if (feature.equals("VMFIN") || feature.equals("PTKNEG") || feature.equals("PRELAT") || feature.equals("PRELS"))
				factor = 25.;
			else if (feature.equals("APPRART") || feature.equals("VVPP") || feature.equals("PTKZU") || feature.equals("PTKVZ") || feature.equals("PWAV")
					|| feature.equals("VVIMP") || feature.equals("PWS"))
				factor = 30.;
			else if (feature.equals("KOKOM") || feature.equals("KOMMA"))
				factor = 40.;
			else if (feature.equals("PTKANT") || feature.equals("PTKA") || feature.equals("PAV") || feature.equals("VAPP") || feature.equals("KOUI")
					|| feature.equals("VAINF") || feature.equals("APZR"))
				factor = 50.;
			else if (feature.equals("SATZINTERN") || feature.equals("PPOSS") || feature.equals("PRF") || feature.equals("VVIZU") || feature.equals("SATZENDE")
					|| feature.equals("ITJ"))
				factor = 100.;
			else {
				// System.err.println("ist POS, aber oben nicht erfasst: " + feature);
				factor = 1.;
			}
		} else if (this.normalizationMode == CALCULATED_MAXIMA_MUSIC) {

			if(feature.equals("8_POS_SET_NOUN_OR_ADJECTIVE"))
				factor = 1.380952380952381;
			else if(feature.equals("8_POS_SET_PERSONAL_OR_POSSESSIV"))
				factor = 4.318181818181818;
			else if(feature.equals("ADJA"))
				factor = 4.142857142857142;
			else if(feature.equals("ADJD"))
				factor = 6.923076923076923;
			else if(feature.equals("ADV"))
				factor = 6.125000000000001;
			else if(feature.equals("APPR"))
				factor = 7.1818181818181825;
			else if(feature.equals("APPRART"))
				factor = 14.5;
			else if(feature.equals("APZR"))
				factor = 57.5;
			else if(feature.equals("ART"))
				factor = 6.416666666666667;
			else if(feature.equals("CARD"))
				factor = 6.410256410256411;
			else if(feature.equals("FM"))
				factor = 6.696969696969697;
			else if(feature.equals("ITJ"))
				factor = 82.0;
			else if(feature.equals("KOKOM"))
				factor = 23.0;
			else if(feature.equals("KON"))
				factor = 2.12;
			else if(feature.equals("KOUI"))
				factor = 61.333333333333336;
			else if(feature.equals("KOUS"))
				factor = 21.333333333333332;
			else if(feature.equals("NE"))
				factor = 2.8666666666666667;
			else if(feature.equals("NN"))
				factor = 3.0526315789473686;
			else if(feature.equals("PAV"))
				factor = 51.333333333333336;
			else if(feature.equals("PDAT"))
				factor = 17.0;
			else if(feature.equals("PDS"))
				factor = 20.833333333333332;
			else if(feature.equals("PIAT"))
				factor = 18.285714285714285;
			else if(feature.equals("PIS"))
				factor = 15.1;
			else if(feature.equals("PPER"))
				factor = 9.5;
			else if(feature.equals("PPOSAT"))
				factor = 13.428571428571429;
			else if(feature.equals("PRELAT"))
				factor = 77.5;
			else if(feature.equals("PRELS"))
				factor = 24.0;
			else if(feature.equals("PRF"))
				factor = 26.750000000000004;
			else if(feature.equals("PTKA"))
				factor = 58.66666666666667;
			else if(feature.equals("PTKANT"))
				factor = 54.5;
			else if(feature.equals("PTKNEG"))
				factor = 22.25;
			else if(feature.equals("PTKVZ"))
				factor = 29.333333333333336;
			else if(feature.equals("PTKZU"))
				factor = 32.33333333333333;
			else if(feature.equals("PWAV"))
				factor = 32.33333333333333;
			else if(feature.equals("PWS"))
				factor = 29.666666666666668;
			else if(feature.equals("TRUNC"))
				factor = 68.5;
			else if(feature.equals("VAFIN"))
				factor = 11.555555555555555;
			else if(feature.equals("VAIMP"))
				factor = 73.66666666666667;
			else if(feature.equals("VAINF"))
				factor = 45.0;
			else if(feature.equals("VAPP"))
				factor = 57.5;
			else if(feature.equals("VMFIN"))
				factor = 24.500000000000004;
			else if(feature.equals("VVFIN"))
				factor = 8.9;
			else if(feature.equals("VVIMP"))
				factor = 36.0;
			else if(feature.equals("VVINF"))
				factor = 16.0;
			else if(feature.equals("VVIZU"))
				factor = 42.5;
			else if(feature.equals("VVPP"))
				factor = 15.0;
			else if(feature.equals("XY"))
				factor = 3.319148936170213;

		} else if (this.normalizationMode == CALCULATED_MAXIMA_SMARTPHONE) {

			if(feature.equals("8_POS_SET_NOUN_OR_ADJECTIVE"))
				factor = 1.2857142857142856;
			else if(feature.equals("8_POS_SET_PERSONAL_OR_POSSESSIV"))
				factor = 4.0;
			else if(feature.equals("ADJA"))
				factor = 3.4;
			else if(feature.equals("ADJD"))
				factor = 6.6;
			else if(feature.equals("ADV"))
				factor = 5.416666666666666;
			else if(feature.equals("APPO"))
				factor = 24.0;
			else if(feature.equals("APPR"))
				factor = 5.666666666666666;
			else if(feature.equals("APPRART"))
				factor = 15.200000000000001;
			else if(feature.equals("APZR"))
				factor = 54.0;
			else if(feature.equals("ART"))
				factor = 5.111111111111111;
			else if(feature.equals("CARD"))
				factor = 7.944444444444444;
			else if(feature.equals("FM"))
				factor = 11.1;
			else if(feature.equals("ITJ"))
				factor = 50.0;
			else if(feature.equals("KOKOM"))
				factor = 15.333333333333334;
			else if(feature.equals("KON"))
				factor = 3.238095238095238;
			else if(feature.equals("KOUI"))
				factor = 29.0;
			else if(feature.equals("KOUS"))
				factor = 21.0;
			else if(feature.equals("NE"))
				factor = 3.042553191489362;
			else if(feature.equals("NN"))
				factor = 2.1666666666666665;
			else if(feature.equals("PDAT"))
				factor = 13.0;
			else if(feature.equals("PDS"))
				factor = 23.5;
			else if(feature.equals("PIAT"))
				factor = 19.0;
			else if(feature.equals("PIS"))
				factor = 15.666666666666668;
			else if(feature.equals("PPER"))
				factor = 8.4;
			else if(feature.equals("PPOSAT"))
				factor = 21.666666666666664;
			else if(feature.equals("PRELAT"))
				factor = 68.0;
			else if(feature.equals("PRELS"))
				factor = 14.0;
			else if(feature.equals("PRF"))
				factor = 20.666666666666668;
			else if(feature.equals("PTKA"))
				factor = 42.0;
			else if(feature.equals("PTKANT"))
				factor = 48.0;
			else if(feature.equals("PTKNEG"))
				factor = 20.5;
			else if(feature.equals("PTKVZ"))
				factor = 17.666666666666668;
			else if(feature.equals("PTKZU"))
				factor = 17.0;
			else if(feature.equals("PWAV"))
				factor = 33.0;
			else if(feature.equals("PWS"))
				factor = 30.0;
			else if(feature.equals("TRUNC"))
				factor = 50.0;
			else if(feature.equals("VAFIN"))
				factor = 11.333333333333332;
			else if(feature.equals("VAIMP"))
				factor = 54.0;
			else if(feature.equals("VAINF"))
				factor = 20.0;
			else if(feature.equals("VAPP"))
				factor = 31.0;
			else if(feature.equals("VMFIN"))
				factor = 18.333333333333336;
			else if(feature.equals("VVFIN"))
				factor = 7.571428571428571;
			else if(feature.equals("VVIMP"))
				factor = 29.5;
			else if(feature.equals("VVINF"))
				factor = 13.0;
			else if(feature.equals("VVIZU"))
				factor = 23.0;
			else if(feature.equals("VVPP"))
				factor = 10.333333333333334;
			else if(feature.equals("XY"))
				factor = 8.7;

		} else if (this.normalizationMode == FeatureHandler.CALCULATED_AVERAGE_MUSIC){
			if(feature.equals("8_POS_SET_NOUN_OR_ADJECTIVE"))
				factor = 0.995577649468709;
			else if(feature.equals("8_POS_SET_PERSONAL_OR_POSSESSIV"))
				factor = 6.591366845796763;
			else if(feature.equals("ADJA"))
				factor = 7.864266772504562;
			else if(feature.equals("ADJD"))
				factor = 13.212499584165528;
			else if(feature.equals("ADV"))
				factor = 7.723312697123633;
			else if(feature.equals("APPR"))
				factor = 9.090769667518515;
			else if(feature.equals("APPRART"))
				factor = 49.18743861931673;
			else if(feature.equals("APZR"))
				factor = 2311.8020085661224;
			else if(feature.equals("ART"))
				factor = 6.927477104997899;
			else if(feature.equals("CARD"))
				factor = 47.39321120008581;
			else if(feature.equals("FM"))
				factor = 44.93238434691419;
			else if(feature.equals("ITJ"))
				factor = 2115.4208136599104;
			else if(feature.equals("KOKOM"))
				factor = 73.86812753439678;
			else if(feature.equals("KON"))
				factor = 17.25187137433181;
			else if(feature.equals("KOUI"))
				factor = 619.8593921596225;
			else if(feature.equals("KOUS"))
				factor = 63.20825092019681;
			else if(feature.equals("NE"))
				factor = 8.116388643398754;
			else if(feature.equals("NN"))
				factor = 2.5195392450193674;
			else if(feature.equals("PAV"))
				factor = 6069.3876316188525;
			else if(feature.equals("PDAT"))
				factor = 98.98271617112488;
			else if(feature.equals("PDS"))
				factor = 133.48374337295078;
			else if(feature.equals("PIAT"))
				factor = 54.299311966301644;
			else if(feature.equals("PIS"))
				factor = 47.85950180070352;
			else if(feature.equals("PPER"))
				factor = 25.304642983227982;
			else if(feature.equals("PPOSAT"))
				factor = 52.433814951752645;
			else if(feature.equals("PRELAT"))
				factor = 2522.7401016425974;
			else if(feature.equals("PRELS"))
				factor = 72.16421161293103;
			else if(feature.equals("PRF"))
				factor = 60.24628833457492;
			else if(feature.equals("PTKA"))
				factor = 969.4415773953341;
			else if(feature.equals("PTKANT"))
				factor = 2073.677542240202;
			else if(feature.equals("PTKNEG"))
				factor = 78.02437851626254;
			else if(feature.equals("PTKVZ"))
				factor = 89.40767671038459;
			else if(feature.equals("PTKZU"))
				factor = 136.26954769713376;
			else if(feature.equals("PWAV"))
				factor = 493.95163338035616;
			else if(feature.equals("PWS"))
				factor = 315.3245718420908;
			else if(feature.equals("TRUNC"))
				factor = 1607.4264941678327;
			else if(feature.equals("VAFIN"))
				factor = 21.53652522244414;
			else if(feature.equals("VAIMP"))
				factor = 3881.6194943412183;
			else if(feature.equals("VAINF"))
				factor = 272.44879801006863;
			else if(feature.equals("VAPP"))
				factor = 996.939888281869;
			else if(feature.equals("VMFIN"))
				factor = 93.97776296421013;
			else if(feature.equals("VVFIN"))
				factor = 11.352246482111584;
			else if(feature.equals("VVIMP"))
				factor = 412.68568172793454;
			else if(feature.equals("VVINF"))
				factor = 45.228446250259594;
			else if(feature.equals("VVIZU"))
				factor = 552.5369708938392;
			else if(feature.equals("VVPP"))
				factor = 42.80814271081066;
			else if(feature.equals("XY"))
				factor = 49.600007112648235;
			else
				System.err.println("Fehler bei POS "+feature);
			
		}else if(this.normalizationMode == FeatureHandler.CALCULATED_AVERAGE_SMARTPHONE){
			if(feature.equals("8_POS_SET_NOUN_OR_ADJECTIVE"))
				factor = 0.9449299902477438;
			else if(feature.equals("8_POS_SET_PERSONAL_OR_POSSESSIV"))
				factor = 8.332049173495347;
			else if(feature.equals("ADJA"))
				factor = 7.947330110933356;
			else if(feature.equals("ADJD"))
				factor = 16.025835981134044;
			else if(feature.equals("ADV"))
				factor = 9.057102197136869;
			else if(feature.equals("APPO"))
				factor = 2526.1524042840274;
			else if(feature.equals("APPR"))
				factor = 8.575078505870122;
			else if(feature.equals("APPRART"))
				factor = 36.96365376222583;
			else if(feature.equals("APZR"))
				factor = 3810.2125470805154;
			else if(feature.equals("ART"))
				factor = 6.096867888379956;
			else if(feature.equals("CARD"))
				factor = 30.947606912611004;
			else if(feature.equals("FM"))
				factor = 174.8031363365384;
			else if(feature.equals("ITJ"))
				factor = 2654.5933532952604;
			else if(feature.equals("KOKOM"))
				factor = 108.99122415967864;
			else if(feature.equals("KON"))
				factor = 19.18573442752603;
			else if(feature.equals("KOUI"))
				factor = 743.0515870839508;
			else if(feature.equals("KOUS"))
				factor = 66.40806012547783;
			else if(feature.equals("NE"))
				factor = 7.567639662490185;
			else if(feature.equals("NN"))
				factor = 2.215852177881805;
			else if(feature.equals("PDAT"))
				factor = 137.7049333559247;
			else if(feature.equals("PDS"))
				factor = 150.3799378404543;
			else if(feature.equals("PIAT"))
				factor = 63.314321271507225;
			else if(feature.equals("PIS"))
				factor = 65.3273935230153;
			else if(feature.equals("PPER"))
				factor = 27.574240698678743;
			else if(feature.equals("PPOSAT"))
				factor = 93.19999627482677;
			else if(feature.equals("PRELAT"))
				factor = 4184.304270666792;
			else if(feature.equals("PRELS"))
				factor = 92.86592881871735;
			else if(feature.equals("PRF"))
				factor = 81.60288539931894;
			else if(feature.equals("PTKA"))
				factor = 1551.8307832499547;
			else if(feature.equals("PTKANT"))
				factor = 2054.9112183497614;
			else if(feature.equals("PTKNEG"))
				factor = 88.29414802976459;
			else if(feature.equals("PTKVZ"))
				factor = 108.34819642792374;
			else if(feature.equals("PTKZU"))
				factor = 134.9728471099148;
			else if(feature.equals("PWAV"))
				factor = 504.6556127529326;
			else if(feature.equals("PWS"))
				factor = 362.54522538855855;
			else if(feature.equals("TRUNC"))
				factor = 2879.448144029765;
			else if(feature.equals("VAFIN"))
				factor = 20.65364258265276;
			else if(feature.equals("VAIMP"))
				factor = 2020.0237619354223;
			else if(feature.equals("VAINF"))
				factor = 161.6964449221814;
			else if(feature.equals("VAPP"))
				factor = 1314.8270660492333;
			else if(feature.equals("VMFIN"))
				factor = 59.84289395374554;
			else if(feature.equals("VVFIN"))
				factor = 11.693994380171185;
			else if(feature.equals("VVIMP"))
				factor = 537.656183765892;
			else if(feature.equals("VVINF"))
				factor = 35.797693288711386;
			else if(feature.equals("VVIZU"))
				factor = 543.000054376622;
			else if(feature.equals("VVPP"))
				factor = 34.505705895247935;
			else if(feature.equals("XY"))
				factor = 173.37684151810979;
		}
		return factor * value / number_of_tokens;
	}

}
